# Copyright (c) 2015 Cisco Systems
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import base64

from barbican.tests import utils
from functionaltests.api import base
from functionaltests.api.v1.behaviors import secret_behaviors
from functionaltests.api.v1.models import secret_models
from functionaltests.common import config


CONF = config.get_config()
admin_a = CONF.rbac_users.admin_a
creator_a = CONF.rbac_users.creator_a
creator_a_2 = CONF.rbac_users.creator_a_2
observer_a = CONF.rbac_users.observer_a
auditor_a = CONF.rbac_users.auditor_a


test_data_rbac_store_secret = {
    'with_admin_a': {'user': admin_a, 'admin': admin_a,
                     'expected_return': 201},
    'with_creator_a': {'user': creator_a, 'admin': admin_a,
                       'expected_return': 201},
    'with_observer_a': {'user': observer_a, 'admin': admin_a,
                        'expected_return': 403},
    'with_auditor_a': {'user': auditor_a, 'admin': admin_a,
                       'expected_return': 403},
}

test_data_rbac_update_secret = {
    'with_admin_a': {'user': admin_a, 'admin': admin_a,
                     'expected_return': 204},
    'with_creator_a': {'user': creator_a, 'admin': admin_a,
                       'expected_return': 204},
    'with_observer_a': {'user': observer_a, 'admin': admin_a,
                        'expected_return': 403},
    'with_auditor_a': {'user': auditor_a, 'admin': admin_a,
                       'expected_return': 403},
}

test_data_rbac_get_secret_metadata = {
    'with_admin_a': {'user': admin_a, 'admin': admin_a,
                     'expected_return': 200},
    'with_creator_a': {'user': creator_a, 'admin': admin_a,
                       'expected_return': 200},
    'with_observer_a': {'user': observer_a, 'admin': admin_a,
                        'expected_return': 200},
    'with_auditor_a': {'user': auditor_a, 'admin': admin_a,
                       'expected_return': 200},
}

test_data_rbac_get_decrypted_secret = {
    'with_admin_a': {'user': admin_a, 'admin': admin_a,
                     'expected_return': 200},
    'with_creator_a': {'user': creator_a, 'admin': admin_a,
                       'expected_return': 200},
    'with_observer_a': {'user': observer_a, 'admin': admin_a,
                        'expected_return': 200},
    'with_auditor_a': {'user': auditor_a, 'admin': admin_a,
                       'expected_return': 403},
}

test_data_rbac_get_list_of_secrets = {
    'with_admin_a': {'user': admin_a, 'admin': admin_a,
                     'expected_return': 200},
    'with_creator_a': {'user': creator_a, 'admin': admin_a,
                       'expected_return': 200},
    'with_observer_a': {'user': observer_a, 'admin': admin_a,
                        'expected_return': 200},
    'with_auditor_a': {'user': auditor_a, 'admin': admin_a,
                       'expected_return': 403},
}

test_data_rbac_delete_secret = {
    'with_admin_a': {'user': admin_a, 'admin': admin_a,
                     'expected_return': 204},
    'with_creator_a': {'user': creator_a, 'admin': creator_a,
                       'expected_return': 204},
    'with_creator_a_2': {'user': creator_a_2, 'admin': creator_a,
                         'expected_return': 403},
    'with_observer_a': {'user': observer_a, 'admin': admin_a,
                        'expected_return': 403},
    'with_auditor_a': {'user': auditor_a, 'admin': admin_a,
                       'expected_return': 403},
}


def get_default_secret_data():
    return {
        "name": "AES key",
        "expiration": "2050-02-28T19:14:44.180394",
        "algorithm": "aes",
        "bit_length": 256,
        "mode": "cbc",
        "payload": get_default_payload(),
        "payload_content_type": get_default_payload_content_type(),
        "payload_content_encoding": get_default_payload_content_encoding(),
    }


def get_default_two_phase_secret_data():
    return {
        "name": "AES key",
        "expiration": "2050-02-28T19:14:44.180394",
        "algorithm": "aes",
        "bit_length": 256,
        "mode": "cbc",
    }


def get_default_two_phase_payload_data():
    return {
        "payload": get_default_payload(),
        "payload_content_type": get_default_payload_content_type(),
        "payload_content_encoding": get_default_payload_content_encoding(),
    }


def get_default_payload():
    return 'Z0Y2K2xMb0Yzb2hBOWFQUnB0KzZiUT09'


def get_default_payload_content_encoding():
    return 'base64'


def get_default_payload_content_type():
    return 'application/octet-stream'


def get_container_req(secret_ref):
    return {"name": "testcontainer",
            "type": "generic",
            "secret_refs": [{'name': 'secret1', 'secret_ref': secret_ref}]}


@utils.parameterized_test_case
class RBACSecretsTestCase(base.TestCase):
    """Functional tests exercising RBAC Policies"""
    def setUp(self):
        super(RBACSecretsTestCase, self).setUp()
        self.secret_behaviors = secret_behaviors.SecretBehaviors(self.client)

    def tearDown(self):
        self.secret_behaviors.delete_all_created_secrets()
        super(RBACSecretsTestCase, self).tearDown()

    @utils.parameterized_dataset(test_data_rbac_store_secret)
    def test_rbac_store_secret(self, user, admin, expected_return):
        """Test RBAC for secret store

        Issue a secret creation and verify that the correct
        http return code comes back for the specified user.

        :param user: the user who will attempt to do the store
        :param admin: the admin of the group containing the user
        :param expected_return: the expected http return code
        """
        test_model = secret_models.SecretModel(**get_default_secret_data())
        resp, secret_ref = self.secret_behaviors.create_secret(
            test_model, user_name=user, admin=admin)
        self.assertEqual(expected_return, resp.status_code)
        self.assertEqual(expected_return == 201, secret_ref is not None)

    @utils.parameterized_dataset(test_data_rbac_update_secret)
    def test_rbac_update_secret(self, user, admin, expected_return):
        """Test RBAC for secret update

        Issue a secret update and verify that the correct
        http return code comes back for the specified user.

        The initial secret will be stored with the admin user to ensure
        that it gets created successfully.  We don't want the secret
        store to fail since we are only testing secret update here.

        :param user: the user who will attempt to do the update
        :param admin: the admin of the group containing the user
        :param expected_return: the expected http return code
        """

        secret_ref = self._create_initial_secret(
            admin=admin, secret_data=get_default_two_phase_secret_data())

        resp = self.secret_behaviors.update_secret_payload(
            secret_ref, user_name=user,
            **get_default_two_phase_payload_data())
        self.assertEqual(expected_return, resp.status_code)

    @utils.parameterized_dataset(test_data_rbac_get_secret_metadata)
    def test_rbac_get_secret_metadata(self, user, admin, expected_return):
        """Test RBAC for secret get metadata

        Issue a secret get metadata and verify that the correct
        http return code comes back for the specified user.

        The initial secret will be stored with the admin user to ensure
        that it gets created successfully.  We don't want the secret
        store to fail since we are only testing secret get metadata here.

        :param user: the user who will attempt to do the get metadata
        :param admin: the admin of the group containing the user
        :param expected_return: the expected http return code
        """
        secret_ref = self._create_initial_secret(admin=admin)

        resp = self.secret_behaviors.get_secret_metadata(
            secret_ref, user_name=user)
        self.assertEqual(expected_return, resp.status_code)
        self.assertEqual(expected_return == 200, resp.content is not None)

    @utils.parameterized_dataset(test_data_rbac_get_decrypted_secret)
    def test_rbac_get_decrypted_secret(self, user, admin, expected_return):
        """Test RBAC for secret get decrypted secret

        Issue a secret get decrypted data and verify that the correct
        http return code comes back for the specified user.

        The initial secret will be stored with the admin user to ensure
        that it gets created successfully.  We don't want the secret
        store to fail since we are only testing get decrypted secret here.

        :param user: the user who will attempt to get the decrypted secret
        :param admin: the admin of the group containing the user
        :param expected_return: the expected http return code
        """
        secret_ref = self._create_initial_secret(admin=admin)

        resp = self.secret_behaviors.get_secret(
            secret_ref, payload_content_type='application/octet-stream',
            user_name=user)
        self.assertEqual(expected_return, resp.status_code)
        self.assertEqual(expected_return == 200,
                         resp.content == base64.b64decode(
                             get_default_payload()))

    @utils.parameterized_dataset(test_data_rbac_get_list_of_secrets)
    def test_rbac_get_list_of_secrets(self, user, admin, expected_return):
        """Test RBAC for get secret list

        Issue a get secret list and verify that the correct
        http return code comes back for the specified user.

        Some initial secrets will be stored with the admin user to ensure
        that they get created successfully.  We don't want the secret
        stores to fail since we are only testing get secret list.

        :param user: the user who will attempt to get the list of secrets
        :param admin: the admin of the group containing the user
        :param expected_return: the expected http return code
        """
        for i in range(3):
            secret_ref = self._create_initial_secret(admin=admin)

        self.assertIsNotNone(secret_ref)
        resp, secrets, next, prev = self.secret_behaviors.get_secrets(
            limit=10, offset=0, user_name=user)
        self.assertEqual(expected_return, resp.status_code)
        self.assertIsNotNone(secrets)

    @utils.parameterized_dataset(test_data_rbac_delete_secret)
    def test_rbac_delete_secret(self, user, admin, expected_return):
        """Test RBAC for secret delete

        Issue a secret delete and verify that the correct
        http return code comes back for the specified user.

        The initial secret will be stored with the admin user to ensure
        that it gets created successfully.  We don't want the secret
        store to fail since we are only testing secret delete here.

        :param user: the user who will attempt to do the delete
        :param admin: the admin of the group containing the user
        :param expected_return: the expected http return code
        """
        secret_ref = self._create_initial_secret(admin=admin)

        resp = self.secret_behaviors.delete_secret(
            secret_ref, user_name=user)
        self.assertEqual(expected_return, resp.status_code)

    def _create_initial_secret(self, admin=admin_a,
                               secret_data=get_default_secret_data()):
        """Utility function to create a secret

        Some tests require a secret to exist before they test certain things,
        so this function can be used to do that setup.

        :param admin: the admin user who will create store the secret
        :param secret_data: the data for the secret
        :return: href to the newly stored secret
        """
        test_model = secret_models.SecretModel(**secret_data)
        resp, secret_ref = self.secret_behaviors.create_secret(
            test_model, user_name=admin, admin=admin)
        self.assertEqual(201, resp.status_code)
        return secret_ref



# Copyright (c) 2015 Rackspace, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import os
from oslo_utils import uuidutils

from barbican.api.controllers import acls
from barbican.model import repositories
from barbican.tests.api import test_resources_policy as test_policy
from barbican.tests import utils


class TestACLsWithContextMixin(test_policy.BaseTestCase):
    """Mixin for performing common acls operation used with policy logic."""

    def _create_secret_with_creator_user(self, app, creator_user_id):

        # define creator user for new secret entry.
        app.extra_environ = {
            'barbican.context': self._build_context(self.project_id,
                                                    user=creator_user_id)
        }
        secret_id, _ = create_secret(app)
        return secret_id

    def _create_container_with_creator_user(self, app, creator_user_id):

        # define creator user for new container entry.
        app.extra_environ = {
            'barbican.context': self._build_context(self.project_id,
                                                    user=creator_user_id)
        }
        container_id, _ = create_container(app)
        return container_id

    def _set_acls_with_context(self, app, entity_type=None, op_type=None,
                               entity_id=None, roles=None, user=None,
                               enforce_policy=True, expect_errors=False):
        """Perform acl create/update/delete operation with policy logic.

        Before performing acl create/update/delete, provided input is used
        for setting custom barbican context. Operation is done under policy
        enforcement logic.
        """
        policy_enforcer = self.policy_enforcer if enforce_policy else None
        app.extra_environ = {
            'barbican.context': self._build_context(
                self.project_id, roles=roles, user=user,
                is_admin=False, policy_enforcer=policy_enforcer)
        }
        resp = None
        if op_type == 'create':
            resp = create_acls(app, entity_type, entity_id,
                               read_user_ids=['u1', 'u2'],
                               expect_errors=expect_errors)
        elif op_type == 'update':
            resp = update_acls(app, entity_type, entity_id,
                               read_user_ids=['u1', 'u2'],
                               partial_update=True,
                               expect_errors=expect_errors)
        elif op_type == 'delete':
            resp = app.delete('/{0}/{1}/acl'.format(entity_type, entity_id),
                              expect_errors=expect_errors)
        return resp


class WhenTestingSecretACLsResource(utils.BarbicanAPIBaseTestCase,
                                    TestACLsWithContextMixin):

    def test_can_create_new_secret_acls(self):
        """Create secret acls and compare stored values with request data."""
        secret_uuid, _ = create_secret(self.app)

        resp = create_acls(
            self.app, 'secrets', secret_uuid,
            read_user_ids=['u1', 'u2'])
        self.assertEqual(200, resp.status_int)
        self.assertIsNotNone(resp.json)
        self.assertIn('/secrets/{0}/acl'.format(secret_uuid),
                      resp.json['acl_ref'])
        acl_map = _get_acl_map(secret_uuid, is_secret=True)
        # Check project_access is True when not provided
        self.assertTrue(acl_map['read']['project_access'])

    def test_who_can_create_new_secret_acls(self):
        """Test who can create new secret ACLs as per policy rules.

        New secret ACLs can be created by user who created the secret.
        Other user with 'creator' role in secret project cannot create ACL
        if user is not creator of the secret.
        User with 'admin' role in secret project can create ACL for that
        secret.
        """
        creator_user_id = 'creatorUserId'
        secret_uuid = self._create_secret_with_creator_user(
            self.app, creator_user_id)
        secret_uuid2 = self._create_secret_with_creator_user(
            self.app, creator_user_id)

        resp = self._set_acls_with_context(
            self.app, entity_type='secrets', op_type='create',
            entity_id=secret_uuid, roles=['creator'], user='NotSecretCreator',
            expect_errors=True)
        self.assertEqual(403, resp.status_int)

        resp = self._set_acls_with_context(
            self.app, entity_type='secrets', op_type='create',
            entity_id=secret_uuid, roles=['creator'],
            user=creator_user_id, expect_errors=False)
        self.assertEqual(200, resp.status_int)

        # test for user with 'admin' role in secret project
        resp = self._set_acls_with_context(
            self.app, entity_type='secrets', op_type='create',
            entity_id=secret_uuid2, roles=['admin'], user='AdminUser',
            expect_errors=False)
        self.assertEqual(200, resp.status_int)

    def test_create_new_secret_acls_with_project_access_false(self):
        """Should allow creating acls for a new secret with project-access."""
        secret_uuid, _ = create_secret(self.app)

        resp = create_acls(
            self.app, 'secrets', secret_uuid,
            read_project_access=False)
        self.assertEqual(200, resp.status_int)
        self.assertIsNotNone(resp.json)
        self.assertIn('/secrets/{0}/acl'.format(secret_uuid),
                      resp.json['acl_ref'])
        acl_map = _get_acl_map(secret_uuid, is_secret=True)
        self.assertFalse(acl_map['read']['project_access'])

    def test_new_secret_acls_with_invalid_project_access_value_should_fail(
            self):
        """Should fail if project-access flag is provided as string value."""
        secret_uuid, _ = create_secret(self.app)

        resp = create_acls(
            self.app, 'secrets', secret_uuid,
            read_project_access="False",
            read_user_ids=['u1', 'u3', 'u4'],
            expect_errors=True)
        self.assertEqual(400, resp.status_int)

        resp = create_acls(
            self.app, 'secrets', secret_uuid,
            read_project_access="None",
            expect_errors=True)
        self.assertEqual(400, resp.status_int)

    def test_get_secret_acls_with_complete_acl_data(self):
        """Read existing acls for a with complete acl data."""
        secret_id, _ = create_secret(self.app)
        create_acls(
            self.app, 'secrets', secret_id,
            read_user_ids=['u1', 'u3'], read_project_access=False)

        resp = self.app.get(
            '/secrets/{0}/acl'.format(secret_id),
            expect_errors=False)
        self.assertEqual(200, resp.status_int)
        self.assertIsNotNone(resp.json)

        self.assertIn('read', resp.json)
        self.assertFalse(resp.json['read']['project-access'])
        self.assertIsNotNone(resp.json['read']['created'])
        self.assertIsNotNone(resp.json['read']['updated'])
        self.assertEqual({'u1', 'u3'}, set(resp.json['read']['users']))

    def test_get_secret_acls_with_project_access_data(self):
        """Read existing acls for acl when only project-access flag is set."""
        secret_id, _ = create_secret(self.app)
        create_acls(
            self.app, 'secrets', secret_id,
            read_project_access=False)

        resp = self.app.get(
            '/secrets/{0}/acl'.format(secret_id),
            expect_errors=False)
        self.assertEqual(200, resp.status_int)
        self.assertIsNotNone(resp.json)

        self.assertEqual([], resp.json['read']['users'])
        self.assertFalse(resp.json['read']['project-access'])
        self.assertIsNotNone(resp.json['read']['created'])
        self.assertIsNotNone(resp.json['read']['updated'])

    def test_get_secret_acls_invalid_secret_should_fail(self):
        """Get secret acls should fail for invalid secret id.

        This test applies to all secret ACLs methods as secret entity is
        populated in same manner for get, put, patch, delete methods.
        """
        secret_id, _ = create_secret(self.app)
        create_acls(
            self.app, 'secrets', secret_id,
            read_project_access=True,
            read_user_ids=['u1', 'u3', 'u4'])

        resp = self.app.get(
            '/secrets/{0}/acl'.format(uuidutils.generate_uuid(dashed=False)),
            expect_errors=True)
        self.assertEqual(404, resp.status_int)

    def test_get_secret_acls_no_acls_defined_return_default_acl(self):
        """Get secret acls should pass when no acls defined for a secret."""
        secret_id, _ = create_secret(self.app)

        resp = self.app.get(
            '/secrets/{0}/acl'.format(secret_id),
            expect_errors=True)
        self.assertEqual(200, resp.status_int)
        self.assertEqual(acls.DEFAULT_ACL, resp.json)

    def test_get_secret_acls_with_incorrect_uri_should_fail(self):
        """Get secret acls should fail when no acls defined for a secret."""
        secret_id, _ = create_secret(self.app)

        resp = self.app.get(
            '/secrets/{0}/incorrect_acls'.format(secret_id),
            expect_errors=True)
        self.assertEqual(404, resp.status_int)

    def test_full_update_secret_acls_modify_project_access_value(self):
        """ACLs full update with userids where project-access flag modified."""
        secret_uuid, _ = create_secret(self.app)

        create_acls(
            self.app, 'secrets', secret_uuid,
            read_user_ids=['u1', 'u2'],
            read_project_access=False)

        # update acls with no user input so  it should delete existing users
        resp = update_acls(
            self.app, 'secrets', secret_uuid, partial_update=False,
            read_project_access=True)
        self.assertEqual(200, resp.status_int)
        self.assertIsNotNone(resp.json)
        self.assertIn('/secrets/{0}/acl'.format(secret_uuid),
                      resp.json['acl_ref'])
        acl_map = _get_acl_map(secret_uuid, is_secret=True)
        self.assertTrue(acl_map['read']['project_access'])
        self.assertIsNone(acl_map['read'].to_dict_fields().get('users'))

    def test_full_update_secret_acls_modify_users_only(self):
        """ACLs full update where specific operation acl is modified."""
        secret_uuid, _ = create_secret(self.app)

        create_acls(
            self.app, 'secrets', secret_uuid,
            read_user_ids=['u1', 'u2'], read_project_access=False)

        resp = update_acls(
            self.app, 'secrets', secret_uuid, partial_update=False,
            read_user_ids=['u1', 'u3', 'u5'])

        self.assertEqual(200, resp.status_int)
        self.assertIsNotNone(resp.json)
        self.assertIn('/secrets/{0}/acl'.format(secret_uuid),
                      resp.json['acl_ref'])
        acl_map = _get_acl_map(secret_uuid, is_secret=True)
        self.assertTrue(acl_map['read']['project_access'])
        self.assertNotIn('u2', acl_map['read'].to_dict_fields()['users'])
        self.assertEqual({'u1', 'u3', 'u5'},
                         set(acl_map['read'].to_dict_fields()['users']))

    def test_full_update_secret_acls_with_read_users_only(self):
        """Acls full update where specific operation acl is modified."""
        secret_uuid, _ = create_secret(self.app)

        create_acls(
            self.app, 'secrets', secret_uuid,
            read_user_ids=['u1', 'u2'])

        acl_map = _get_acl_map(secret_uuid, is_secret=True)
        # ACL api does not support 'list' operation so making direct db update
        # in acl operation data to make sure full update removes this existing
        # ACL.
        secret_acl = acl_map['read']
        secret_acl.operation = 'list'
        secret_acl.save()
        acl_map = _get_acl_map(secret_uuid, is_secret=True)
        # check 'list' operation is there in db
        self.assertIn('list', acl_map)
        resp = update_acls(
            self.app, 'secrets', secret_uuid, partial_update=False,
            read_user_ids=['u1', 'u3', 'u5'])

        self.assertEqual(200, resp.status_int)
        self.assertIsNotNone(resp.json)
        self.assertIn('/secrets/{0}/acl'.format(secret_uuid),
                      resp.json['acl_ref'])
        acl_map = _get_acl_map(secret_uuid, is_secret=True)
        # make sure 'list' operation is no longer after full update
        self.assertNotIn('list', acl_map)
        self.assertTrue(acl_map['read']['project_access'])
        self.assertEqual({'u1', 'u3', 'u5'},
                         set(acl_map['read'].to_dict_fields()['users']))
        self.assertNotIn('u2', acl_map['read'].to_dict_fields()['users'])

    def test_partial_update_secret_acls_with_read_users_only(self):
        """Acls update where specific operation acl is modified."""
        secret_uuid, _ = create_secret(self.app)

        create_acls(
            self.app, 'secrets', secret_uuid,
            read_user_ids=['u1', 'u2'])

        acl_map = _get_acl_map(secret_uuid, is_secret=True)

        secret_acl = acl_map['read']
        secret_acl.operation = 'list'
        secret_acl.save()
        acl_map = _get_acl_map(secret_uuid, is_secret=True)
        # check 'list' operation is there in db
        self.assertIn('list', acl_map)
        resp = update_acls(
            self.app, 'secrets', secret_uuid, partial_update=True,
            read_user_ids=['u1', 'u3', 'u5'])

        self.assertEqual(200, resp.status_int)
        self.assertIsNotNone(resp.json)
        self.assertIn('/secrets/{0}/acl'.format(secret_uuid),
                      resp.json['acl_ref'])
        acl_map = _get_acl_map(secret_uuid, is_secret=True)
        # For partial update, existing other operation ACL is not tocuhed.
        self.assertIn('list', acl_map)
        self.assertEqual({'u1', 'u2'},
                         set(acl_map['list'].to_dict_fields()['users']))
        self.assertTrue(acl_map['read']['project_access'])
        self.assertEqual({'u1', 'u3', 'u5'},
                         set(acl_map['read'].to_dict_fields()['users']))

    def test_partial_update_secret_acls_when_no_acls_defined_should_pass(self):
        """Acls partial update pass when no acls are defined for a secret.

        Partial update (PATCH) is applicable even when no explicit ACL has been
        set as by default every secret has implicit acl definition. If PUT
        is used, then new ACL is created instead.
        """
        secret_id, _ = create_secret(self.app)

        resp = update_acls(
            self.app, 'secrets', secret_id, partial_update=True,
            read_user_ids=['u1', 'u3', 'u5'], expect_errors=False)

        self.assertEqual(200, resp.status_int)
        acl_map = _get_acl_map(secret_id, is_secret=True)
        self.assertTrue(acl_map['read']['project_access'])

    def test_who_can_update_secret_acls(self):
        """Test PATCH update existing secret ACLs as per policy rules.

        Existing secret ACLs can be updated by user who created the secret.
        Other user with 'creator' role in secret project cannot update ACL
        if user is not creator of the secret.
        User with 'admin' role in secret project can update ACL for that
        secret.
        """
        creator_user_id = 'creatorUserId'
        secret_uuid = self._create_secret_with_creator_user(
            self.app, creator_user_id)

        self._set_acls_with_context(
            self.app, entity_type='secrets', op_type='create',
            entity_id=secret_uuid, enforce_policy=False)

        resp = self._set_acls_with_context(
            self.app, entity_type='secrets', op_type='update',
            entity_id=secret_uuid, roles=['creator'], user='NotSecretCreator',
            expect_errors=True)
        self.assertEqual(403, resp.status_int)

        resp = self._set_acls_with_context(
            self.app, entity_type='secrets', op_type='update',
            entity_id=secret_uuid, roles=['creator'],
            user=creator_user_id)
        self.assertEqual(200, resp.status_int)

        # test for user with 'admin' role in secret project
        resp = self._set_acls_with_context(
            self.app, entity_type='secrets', op_type='update',
            entity_id=secret_uuid, roles=['admin'], user='AdminUser')
        self.assertEqual(200, resp.status_int)

    def test_partial_update_secret_acls_modify_project_access_values(self):
        """Acls partial update where project-access flag is modified."""
        secret_uuid, _ = create_secret(self.app)

        create_acls(
            self.app, 'secrets', secret_uuid,
            read_user_ids=['u1', 'u2'],
            read_project_access=False)

        resp = update_acls(
            self.app, 'secrets', secret_uuid, partial_update=True,
            read_project_access=True)
        self.assertEqual(200, resp.status_int)
        self.assertIsNotNone(resp.json)
        self.assertIn('/secrets/{0}/acl'.format(secret_uuid),
                      resp.json['acl_ref'])
        acl_map = _get_acl_map(secret_uuid, is_secret=True)
        self.assertTrue(acl_map['read']['project_access'])
        self.assertEqual({'u1', 'u2'},
                         set(acl_map['read'].to_dict_fields()['users']))

    def test_delete_secret_acls_with_valid_secret_id(self):
        """Delete existing acls for a given secret."""
        secret_id, _ = create_secret(self.app)
        create_acls(
            self.app, 'secrets', secret_id,
            read_project_access=True)

        resp = self.app.delete(
            '/secrets/{0}/acl'.format(secret_id),
            expect_errors=False)
        content = resp.json
        self.assertIsNone(content)  # make sure there is no response
        self.assertEqual(200, resp.status_int)
        acl_map = _get_acl_map(secret_id, is_secret=True)
        self.assertFalse(acl_map)

    def test_delete_secret_acls_no_acl_defined_should_pass(self):
        """Delete acls should pass when no acls are defined for a secret."""
        secret_id, _ = create_secret(self.app)

        resp = self.app.delete(
            '/secrets/{0}/acl'.format(secret_id),
            expect_errors=False)
        self.assertEqual(200, resp.status_int)

    def test_who_can_delete_secret_acls(self):
        """Test who can delete existing secret ACLs as per policy rules.

        Existing secret ACLs can be deleted by user who created the secret.
        Other user with 'creator' role in secret project cannot delete ACL
        if user is not creator of the secret.
        User with 'admin' role in secret project can delete ACL for that
        secret.
        """
        creator_user_id = 'creatorUserId'
        secret_uuid = self._create_secret_with_creator_user(
            self.app, creator_user_id)

        self._set_acls_with_context(
            self.app, entity_type='secrets', op_type='create',
            entity_id=secret_uuid, enforce_policy=False)

        resp = self._set_acls_with_context(
            self.app, entity_type='secrets', op_type='delete',
            entity_id=secret_uuid, roles=['creator'], user='NotSecretCreator',
            expect_errors=True)

        self.assertEqual(403, resp.status_int)

        resp = self._set_acls_with_context(
            self.app, entity_type='secrets', op_type='delete',
            entity_id=secret_uuid, roles=['creator'],
            user=creator_user_id)

        self.assertEqual(200, resp.status_int)

        # Create new secret ACLs again.
        self._set_acls_with_context(
            self.app, entity_type='secrets', op_type='create',
            entity_id=secret_uuid, enforce_policy=False)
        # test for user with 'admin' role in secret project
        resp = self._set_acls_with_context(
            self.app, entity_type='secrets', op_type='delete',
            entity_id=secret_uuid, roles=['admin'],
            user='AdminUser')
        self.assertEqual(200, resp.status_int)

    def test_invoke_secret_acls_head_should_fail(self):
        """Should fail as put request to secret acls URI is not supported."""
        secret_id, _ = create_secret(self.app)
        resp = self.app.head(
            '/secrets/{0}/acl'.format(secret_id),
            expect_errors=True)
        self.assertEqual(405, resp.status_int)

    def test_list_secrets_with_no_acls_and_acl_only_should_be_empty(self):
        """Return list should be empty"""
        creator_user_id = 'creatorUserID'
        self._create_secret_with_creator_user(
            self.app, creator_user_id)

        resp = self.app.get(
            '/secrets/?acl_only=TRUE')
        self.assertEqual(200, resp.status_int)
        self.assertEqual([], resp.json['secrets'])

    def test_list_secrets_with_acls(self):
        """Return List should not include secrets with no ACL for user"""
        creator_user_id = 'creatorUserID'
        secret_uuid_acl_1 = self._create_secret_with_creator_user(
            self.app, creator_user_id)
        secret_uuid_acl_2 = self._create_secret_with_creator_user(
            self.app, creator_user_id)
        secret_uuid_no_acl = self._create_secret_with_creator_user(
            self.app, creator_user_id)
        create_acls(
            self.app, 'secrets', secret_uuid_acl_1,
            read_user_ids=[creator_user_id],
            read_project_access=False)

        create_acls(
            self.app, 'secrets', secret_uuid_acl_2,
            read_user_ids=[creator_user_id],
            read_project_access=False)
        resp = self.app.get(
            '/secrets/?acl_only=TrUe')
        self.assertEqual(200, resp.status_int)
        secret_list = resp.json.get('secrets')
        self.assertEqual(2, len(secret_list))

        self.assertNotIn(secret_uuid_no_acl, secret_list)


class WhenTestingContainerAclsResource(utils.BarbicanAPIBaseTestCase,
                                       TestACLsWithContextMixin):

    def test_can_create_new_container_acls(self):
        """Create container acls and compare db values with request data."""
        container_id, _ = create_container(self.app)

        resp = create_acls(
            self.app, 'containers', container_id,
            read_user_ids=['u1', 'u2'])
        self.assertEqual(200, resp.status_int)
        self.assertIsNotNone(resp.json)
        self.assertIn('/containers/{0}/acl'.format(container_id),
                      resp.json['acl_ref'])

        acl_map = _get_acl_map(container_id, is_secret=False)
        # Check project_access is True when not provided
        self.assertTrue(acl_map['read']['project_access'])
        self.assertEqual({'u1', 'u2'},
                         set(acl_map['read'].to_dict_fields()['users']))

    def test_who_can_create_new_container_acls(self):
        """Test who can create new container ACLs as per policy rules.

        New container ACLs can be created by user who created the container.
        Other user with 'creator' role in container project cannot create ACL
        if user is not creator of the container.
        User with 'admin' role in container project can create ACL for that
        container.
        """
        creator_user_id = 'creatorUserId'
        container_id = self._create_container_with_creator_user(
            self.app, creator_user_id)
        container_id2 = self._create_container_with_creator_user(
            self.app, creator_user_id)

        resp = self._set_acls_with_context(
            self.app, entity_type='containers', op_type='create',
            entity_id=container_id, roles=['creator'],
            user='NotContainerCreator', expect_errors=True)
        self.assertEqual(403, resp.status_int)

        resp = self._set_acls_with_context(
            self.app, entity_type='containers', op_type='create',
            entity_id=container_id, roles=['creator'],
            user=creator_user_id, expect_errors=False)
        self.assertEqual(200, resp.status_int)

        # test for user with 'admin' role in container project
        resp = self._set_acls_with_context(
            self.app, entity_type='containers', op_type='create',
            entity_id=container_id2, roles=['admin'], user='AdminUser',
            expect_errors=False)
        self.assertEqual(200, resp.status_int)

    def test_create_new_container_acls_with_project_access_true(self):
        """Should allow creating acls for new container with project-access."""
        container_id, _ = create_container(self.app)

        resp = create_acls(
            self.app, 'containers', container_id,
            read_project_access=True,
            read_user_ids=['u1', 'u3', 'u4'])
        self.assertEqual(200, resp.status_int)
        self.assertIsNotNone(resp.json)
        self.assertIn('/containers/{0}/acl'.format(container_id),
                      resp.json['acl_ref'])
        acl_map = _get_acl_map(container_id, is_secret=False)
        self.assertTrue(acl_map['read']['project_access'])

    def test_create_new_container_acls_with_project_access_false(self):
        """Should allow creating acls for new container with project-access."""
        container_id, _ = create_container(self.app)

        resp = create_acls(
            self.app, 'containers', container_id,
            read_project_access=False,
            read_user_ids=['u1', 'u3', 'u4'])
        self.assertEqual(200, resp.status_int)
        self.assertIsNotNone(resp.json)
        self.assertIn('/containers/{0}/acl'.format(container_id),
                      resp.json['acl_ref'])
        acl_map = _get_acl_map(container_id, is_secret=False)
        self.assertFalse(acl_map['read']['project_access'])

    def test_container_acls_with_invalid_project_access_value_fail(self):
        """Should fail if project-access flag is provided as string value."""
        container_id, _ = create_container(self.app)

        resp = create_acls(
            self.app, 'containers', container_id,
            read_project_access="False",
            read_user_ids=['u1', 'u3', 'u4'],
            expect_errors=True)
        self.assertEqual(400, resp.status_int)

        resp = create_acls(
            self.app, 'containers', container_id,
            read_project_access="None",
            expect_errors=True)
        self.assertEqual(400, resp.status_int)

    def test_get_container_acls_with_complete_acl_data(self):
        """Read existing acls for a with complete acl data."""
        container_id, _ = create_container(self.app)
        create_acls(
            self.app, 'containers', container_id,
            read_user_ids=['u1', 'u3'], read_project_access=False)

        resp = self.app.get(
            '/containers/{0}/acl'.format(container_id),
            expect_errors=False)
        self.assertEqual(200, resp.status_int)
        self.assertIsNotNone(resp.json)

        self.assertIn('read', resp.json)
        self.assertFalse(resp.json['read']['project-access'])
        self.assertIsNotNone(resp.json['read']['created'])
        self.assertIsNotNone(resp.json['read']['updated'])
        self.assertEqual({'u1', 'u3'}, set(resp.json['read']['users']))

    def test_get_container_acls_with_project_access_data(self):
        """Read existing acls for acl when only project-access flag is set."""
        container_id, _ = create_container(self.app)
        create_acls(
            self.app, 'containers', container_id,
            read_project_access=False)

        resp = self.app.get(
            '/containers/{0}/acl'.format(container_id),
            expect_errors=False)
        self.assertEqual(200, resp.status_int)
        self.assertIsNotNone(resp.json)

        self.assertEqual([], resp.json['read']['users'])
        self.assertFalse(resp.json['read']['project-access'])
        self.assertIsNotNone(resp.json['read']['created'])
        self.assertIsNotNone(resp.json['read']['updated'])

    def test_get_container_acls_invalid_container_id_should_fail(self):
        """Get container acls should fail for invalid secret id.

        This test applies to all container ACLs methods as secret entity is
        populated in same manner for get, put, patch, delete methods.
        """
        container_id, _ = create_container(self.app)
        create_acls(
            self.app, 'containers', container_id,
            read_project_access=True)

        resp = self.app.get(
            '/containers/{0}/acl'.format(uuidutils.generate_uuid()),
            expect_errors=True)
        self.assertEqual(404, resp.status_int)

    def test_get_container_acls_invalid_non_uuid_secret_should_fail(self):
        """Get container acls should fail for invalid (non-uuid) id."""
        container_id, _ = create_container(self.app)
        create_acls(
            self.app, 'containers', container_id,
            read_project_access=True)

        resp = self.app.get(
            '/containers/{0}/acl'.format('my_container_id'),
            expect_errors=True)
        self.assertEqual(404, resp.status_int)

    def test_get_container_acls_no_acls_defined_return_default_acl(self):
        """Get container acls should pass when no acls defined for a secret."""
        container_id, _ = create_container(self.app)

        resp = self.app.get(
            '/containers/{0}/acl'.format(container_id),
            expect_errors=True)
        self.assertEqual(200, resp.status_int)
        self.assertEqual(acls.DEFAULT_ACL, resp.json)

    def test_full_update_container_acls_modify_all_acls(self):
        """Acls update where only user ids list is modified."""
        container_id, _ = create_container(self.app)

        create_acls(
            self.app, 'containers', container_id, read_project_access=False,
            read_user_ids=['u1', 'u2'])

        resp = update_acls(
            self.app, 'containers', container_id, partial_update=False,
            read_user_ids=['u1', 'u2', 'u5'])

        self.assertEqual(200, resp.status_int)
        self.assertIsNotNone(resp.json)
        self.assertIn('/containers/{0}/acl'.format(container_id),
                      resp.json['acl_ref'])
        acl_map = _get_acl_map(container_id, is_secret=False)
        # Check project_access is True when not provided
        self.assertTrue(acl_map['read']['project_access'])
        self.assertIn('u5', acl_map['read'].to_dict_fields()['users'])

    def test_full_update_container_acls_modify_project_access_values(self):
        """Acls update where user ids and project-access flag is modified."""
        container_id, _ = create_container(self.app)

        create_acls(
            self.app, 'containers', container_id,
            read_user_ids=['u1', 'u2'])

        resp = update_acls(
            self.app, 'containers', container_id, partial_update=False,
            read_project_access=False)
        self.assertEqual(200, resp.status_int)
        self.assertIsNotNone(resp.json)
        self.assertIn('/containers/{0}/acl'.format(container_id),
                      resp.json['acl_ref'])
        acl_map = _get_acl_map(container_id, is_secret=False)
        self.assertFalse(acl_map['read']['project_access'])
        self.assertIsNone(acl_map['read'].to_dict_fields().get('users'))

    def test_full_update_container_acls_with_read_users_only(self):
        """Acls full update where specific operation acl is modified."""
        container_id, _ = create_container(self.app)

        create_acls(
            self.app, 'containers', container_id,
            read_user_ids=['u1', 'u2'])

        acl_map = _get_acl_map(container_id, is_secret=False)
        # ACL api does not support 'list' operation so making direct db update
        # in acl operation data to make sure full update removes this existing
        # ACL.
        container_acl = acl_map['read']
        container_acl.operation = 'list'
        container_acl.save()
        acl_map = _get_acl_map(container_id, is_secret=False)
        # check 'list' operation is there in db
        self.assertIn('list', acl_map)
        resp = update_acls(
            self.app, 'containers', container_id, partial_update=False,
            read_user_ids=['u1', 'u3', 'u5'])

        self.assertEqual(200, resp.status_int)
        self.assertIsNotNone(resp.json)
        self.assertIn('/containers/{0}/acl'.format(container_id),
                      resp.json['acl_ref'])
        acl_map = _get_acl_map(container_id, is_secret=False)
        # make sure 'list' operation is no longer after full update
        self.assertNotIn('list', acl_map)
        self.assertTrue(acl_map['read']['project_access'])
        self.assertEqual({'u1', 'u3', 'u5'},
                         set(acl_map['read'].to_dict_fields()['users']))
        self.assertNotIn('u2', acl_map['read'].to_dict_fields()['users'])

    def test_partial_update_container_acls_with_read_users_only(self):
        """Acls update where specific operation acl is modified."""
        container_id, _ = create_container(self.app)

        create_acls(
            self.app, 'containers', container_id,
            read_user_ids=['u1', 'u2'])

        acl_map = _get_acl_map(container_id, is_secret=False)

        secret_acl = acl_map['read']
        secret_acl.operation = 'list'
        secret_acl.save()
        acl_map = _get_acl_map(container_id, is_secret=False)
        # check 'list' operation is there in db
        self.assertIn('list', acl_map)
        resp = update_acls(
            self.app, 'containers', container_id, partial_update=True,
            read_user_ids=['u1', 'u3', 'u5'])

        self.assertEqual(200, resp.status_int)
        self.assertIsNotNone(resp.json)
        self.assertIn('/containers/{0}/acl'.format(container_id),
                      resp.json['acl_ref'])
        acl_map = _get_acl_map(container_id, is_secret=False)
        # For partial update, existing other operation ACL is not tocuhed.
        self.assertIn('list', acl_map)
        self.assertEqual({'u1', 'u2'},
                         set(acl_map['list'].to_dict_fields()['users']))
        self.assertTrue(acl_map['read']['project_access'])
        self.assertEqual({'u1', 'u3', 'u5'},
                         set(acl_map['read'].to_dict_fields()['users']))

    def test_partial_update_container_acls_when_no_acls_defined(self):
        """Acls partial update pass when no acls are defined for container.

        Partial update (PATCH) is applicable even when no explicit ACL has been
        set as by default every container has implicit acl definition. If PUT
        is used, then new ACL is created instead.
        """
        container_id, _ = create_container(self.app)

        resp = update_acls(
            self.app, 'containers', container_id, partial_update=True,
            read_user_ids=['u1', 'u3', 'u5'], expect_errors=False)

        self.assertEqual(200, resp.status_int)
        acl_map = _get_acl_map(container_id, is_secret=False)
        self.assertTrue(acl_map['read']['project_access'])

    def test_partial_update_container_acls_modify_project_access_values(self):
        """Acls partial update where project-access flag is modified."""
        container_id, _ = create_container(self.app)

        create_acls(
            self.app, 'containers', container_id,
            read_user_ids=['u1', 'u2'],
            read_project_access=False)

        resp = update_acls(
            self.app, 'containers', container_id, partial_update=True,
            read_project_access=True)
        self.assertEqual(200, resp.status_int)
        self.assertIsNotNone(resp.json)
        self.assertIn('/containers/{0}/acl'.format(container_id),
                      resp.json['acl_ref'])
        acl_map = _get_acl_map(container_id, is_secret=False)
        self.assertTrue(acl_map['read']['project_access'])
        self.assertEqual({'u1', 'u2'},
                         set(acl_map['read'].to_dict_fields()['users']))

    def test_who_can_update_container_acls(self):
        """Test PATCH update existing container ACLs as per policy rules.

        Existing container ACLs can be updated by user who created the
        container.
        Other user with 'creator' role in container project cannot update ACL
        if user is not creator of the container.
        User with 'admin' role in container project can update ACL for that
        container.
        """
        creator_user_id = 'creatorUserId'
        container_id = self._create_container_with_creator_user(
            self.app, creator_user_id)

        self._set_acls_with_context(
            self.app, entity_type='containers', op_type='create',
            entity_id=container_id, enforce_policy=False)

        resp = self._set_acls_with_context(
            self.app, entity_type='containers', op_type='update',
            entity_id=container_id, roles=['creator'], user='NotCreator',
            expect_errors=True)
        self.assertEqual(403, resp.status_int)

        resp = self._set_acls_with_context(
            self.app, entity_type='containers', op_type='update',
            entity_id=container_id, roles=['creator'],
            user=creator_user_id)
        self.assertEqual(200, resp.status_int)

        # test for user with 'admin' role in container project
        resp = self._set_acls_with_context(
            self.app, entity_type='containers', op_type='update',
            entity_id=container_id, roles=['admin'], user='AdminUser')
        self.assertEqual(200, resp.status_int)

    def test_delete_container_acls_with_valid_container_id(self):
        """Delete existing acls for a given container."""
        container_id, _ = create_container(self.app)
        create_acls(
            self.app, 'containers', container_id,
            read_project_access=True)

        resp = self.app.delete(
            '/containers/{0}/acl'.format(container_id),
            expect_errors=False)
        content = resp.json
        self.assertIsNone(content)  # make sure there is no response
        self.assertEqual(200, resp.status_int)
        acl_map = _get_acl_map(container_id, is_secret=False)
        self.assertFalse(acl_map)

    def test_delete_container_acls_no_acl_defined_should_pass(self):
        """Delete acls should pass when no acls are defined for a container."""
        container_id, _ = create_container(self.app)
        resp = self.app.delete(
            '/containers/{0}/acl'.format(container_id),
            expect_errors=False)
        self.assertEqual(200, resp.status_int)

    def test_who_can_delete_container_acls(self):
        """Test who can delete existing container ACLs as per policy rules.

        Existing container ACLs can be deleted by user who created the
        container.
        Other user with 'creator' role in container project cannot delete ACL
        if user is not creator of the container.
        User with 'admin' role in container project can delete ACL for that
        container.
        """
        creator_user_id = 'creatorUserId'
        container_id = self._create_container_with_creator_user(
            self.app, creator_user_id)

        self._set_acls_with_context(
            self.app, entity_type='containers', op_type='create',
            entity_id=container_id, enforce_policy=False)

        resp = self._set_acls_with_context(
            self.app, entity_type='containers', op_type='delete',
            entity_id=container_id, roles=['creator'], user='NotCreator',
            expect_errors=True)

        self.assertEqual(403, resp.status_int)

        resp = self._set_acls_with_context(
            self.app, entity_type='containers', op_type='delete',
            entity_id=container_id, roles=['creator'],
            user=creator_user_id)

        self.assertEqual(200, resp.status_int)

        # Create new container ACLs again.
        self._set_acls_with_context(
            self.app, entity_type='containers', op_type='create',
            entity_id=container_id, enforce_policy=False)
        # test for user with 'admin' role in container project
        resp = self._set_acls_with_context(
            self.app, entity_type='containers', op_type='delete',
            entity_id=container_id, roles=['admin'],
            user='AdminUser')
        self.assertEqual(200, resp.status_int)

    def test_invoke_container_acls_head_should_fail(self):
        """PUT request to container acls URI is not supported."""
        container_id, _ = create_container(self.app)
        resp = self.app.head(
            '/containers/{0}/acl/'.format(container_id),
            expect_errors=True)
        self.assertEqual(405, resp.status_int)


# ----------------------- Helper Functions ---------------------------
def create_secret(app, name=None, algorithm=None, bit_length=None, mode=None,
                  expiration=None, payload='not-encrypted',
                  content_type='text/plain',
                  content_encoding=None, transport_key_id=None,
                  transport_key_needed=None, expect_errors=False):
    request = {
        'name': name,
        'algorithm': algorithm,
        'bit_length': bit_length,
        'mode': mode,
        'expiration': expiration,
        'payload': payload,
        'payload_content_type': content_type,
        'payload_content_encoding': content_encoding,
        'transport_key_id': transport_key_id,
        'transport_key_needed': transport_key_needed
    }
    cleaned_request = {key: val for key, val in request.items()
                       if val is not None}

    resp = app.post_json(
        '/secrets/',
        cleaned_request,
        expect_errors=expect_errors
    )
    created_uuid = None
    if resp.status_int == 201:
        secret_ref = resp.json.get('secret_ref', '')
        _, created_uuid = os.path.split(secret_ref)

    return created_uuid, resp


def create_container(app):
    _, resp = create_secret(app)
    secret_ref = resp.json['secret_ref']
    request = {
        "name": "container name",
        "type": "generic",
        "secret_refs": [
            {
                "name": "any_key",
                "secret_ref": secret_ref
            }
        ]
    }
    resp = app.post_json(
        '/containers/',
        request,
        expect_errors=False
    )
    created_uuid = None
    if resp.status_int == 201:
        container_ref = resp.json.get('container_ref', '')
        _, created_uuid = os.path.split(container_ref)

    return created_uuid, resp


def create_acls(app, entity_type, entity_id, read_user_ids=None,
                read_project_access=None,
                expect_errors=False):
    return manage_acls(app, entity_type, entity_id,
                       read_user_ids=read_user_ids,
                       read_project_access=read_project_access,
                       is_update=False, partial_update=False,
                       expect_errors=expect_errors)


def update_acls(app, entity_type, entity_id, read_user_ids=None,
                read_project_access=None, partial_update=False,
                expect_errors=False):
    return manage_acls(app, entity_type, entity_id,
                       read_user_ids=read_user_ids,
                       read_project_access=read_project_access,
                       is_update=True, partial_update=partial_update,
                       expect_errors=expect_errors)


def manage_acls(app, entity_type, entity_id, read_user_ids=None,
                read_project_access=None, is_update=False,
                partial_update=None, expect_errors=False):
    request = {}

    _append_acl_to_request(request, 'read', read_user_ids,
                           read_project_access)

    cleaned_request = {key: val for key, val in request.items()
                       if val is not None}

    if is_update and partial_update:  # patch for partial update
        resp = app.patch_json(
            '/{0}/{1}/acl'.format(entity_type, entity_id),
            cleaned_request,
            expect_errors=expect_errors)
    else:  # put (for create or complete update)
        resp = app.put_json(
            '/{0}/{1}/acl'.format(entity_type, entity_id),
            cleaned_request,
            expect_errors=expect_errors)

    return resp


def _append_acl_to_request(req, operation, user_ids=None, project_access=None):
    op_dict = {}
    if user_ids is not None:
        op_dict['users'] = user_ids
    if project_access is not None:
        op_dict['project-access'] = project_access
    if op_dict:
        req[operation] = op_dict


def _get_acl_map(entity_id, is_secret=True):
    """Provides map of operation: acl_entity for given entity id."""
    if is_secret:
        acl_repo = repositories.get_secret_acl_repository()
        acl_map = {acl.operation: acl for acl in
                   acl_repo.get_by_secret_id(entity_id)}
    else:
        acl_repo = repositories.get_container_acl_repository()
        acl_map = {acl.operation: acl for acl in
                   acl_repo.get_by_container_id(entity_id)}
    return acl_map



# Copyright (c) 2015 Rackspace, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import base64
import os
from unittest import mock

from oslo_utils import timeutils

from barbican.api.controllers import secrets
from barbican.common import validators
from barbican.model import models
from barbican.model import repositories
from barbican.tests import utils

project_repo = repositories.get_project_repository()
secrets_repo = repositories.get_secret_repository()
tkey_repo = repositories.get_transport_key_repository()


@utils.parameterized_test_case
class WhenTestingSecretsResource(utils.BarbicanAPIBaseTestCase):

    def test_can_create_new_secret_one_step(self):
        resp, secret_uuid = create_secret(
            self.app,
            payload='not-encrypted',
            content_type='text/plain'
        )

        self.assertEqual(201, resp.status_int)
        self.assertIsNotNone(secret_uuid)

    def test_can_create_new_secret_without_payload(self):
        resp, secret_uuid = create_secret(self.app, name='test')
        self.assertEqual(201, resp.status_int)

        secret = secrets_repo.get(secret_uuid, self.project_id)
        self.assertEqual('test', secret.name)
        self.assertEqual([], secret.encrypted_data)

    def test_can_create_new_secret_if_project_doesnt_exist(self):
        # Build new context
        new_project_context = self._build_context('test_project_id')
        self.app.extra_environ = {'barbican.context': new_project_context}

        # Create a generic secret
        resp, _ = create_secret(self.app, name='test_secret')
        self.assertEqual(201, resp.status_int)

        # Verify the new project was created
        project = project_repo.find_by_external_project_id('test_project_id')
        self.assertIsNotNone(project)

    def test_can_create_new_secret_with_payload_just_under_max(self):
        large_payload = 'A' * (validators.DEFAULT_MAX_SECRET_BYTES - 8)
        resp, _ = create_secret(
            self.app,
            payload=large_payload,
            content_type='text/plain'
        )

        self.assertEqual(201, resp.status_int)

    def test_creating_new_secret_with_oversized_payload_should_fail(self):
        oversized_payload = 'A' * (validators.DEFAULT_MAX_SECRET_BYTES + 10)
        resp, _ = create_secret(
            self.app,
            payload=oversized_payload,
            content_type='text/plain',
            expect_errors=True
        )
        self.assertEqual(413, resp.status_int)

    def test_create_new_secret_with_empty_payload_should_fail(self):
        resp, _ = create_secret(
            self.app,
            payload='',
            content_type='text/plain',
            expect_errors=True
        )
        self.assertEqual(400, resp.status_int)

    def test_expiration_should_be_normalized_with_new_secret(self):
        target_expiration = '2114-02-28 12:14:44.180394-05:00'
        resp, secret_uuid = create_secret(
            self.app,
            expiration=target_expiration
        )

        self.assertEqual(201, resp.status_int)

        # Verify that the system normalizes time to UTC
        secret = secrets_repo.get(secret_uuid, self.project_id)
        local_datetime = timeutils.parse_isotime(target_expiration)
        datetime_utc = timeutils.normalize_time(local_datetime)

        self.assertEqual(datetime_utc, secret.expiration)

    @mock.patch('barbican.plugin.resources.store_secret')
    def test_can_create_new_secret_meta_w_transport_key(self, mocked_store):
        transport_key_model = models.TransportKey('default_plugin', 'tkey1234')

        # TODO(jvrbanac): Look into removing this patch
        mocked_store.return_value = models.Secret(), transport_key_model

        # Make sure to add the transport key
        tkey_repo.create_from(transport_key_model)
        transport_key_id = transport_key_model.id

        resp, secret_uuid = create_secret(
            self.app,
            name='test',
            transport_key_needed='true'
        )
        self.assertEqual(201, resp.status_int)
        self.assertIsNotNone(secret_uuid)
        self.assertIn(transport_key_id, resp.json.get('transport_key_ref'))

    @mock.patch('barbican.plugin.resources.store_secret')
    def test_can_create_new_secret_with_transport_key(self, mocked_store):
        # TODO(jvrbanac): Look into removing this patch
        mocked_store.return_value = models.Secret(), None

        # Create Transport Key (keeping for session scoping reasons)
        transport_key_model = models.TransportKey('default_plugin', 'tkey1234')
        transport_key_id = transport_key_model.id
        tkey_repo.create_from(transport_key_model)

        # Create a normal secret with the TransportKey
        resp, secret_uuid = create_secret(
            self.app,
            payload='not-encrypted',
            content_type='text/plain',
            transport_key_id=transport_key_id
        )

        self.assertEqual(201, resp.status_int)
        # We're interested in the transport key values
        mocked_store.assert_called_once_with(
            unencrypted_raw='not-encrypted',
            content_type_raw='text/plain',
            content_encoding=None,
            secret_model=mock.ANY,
            project_model=mock.ANY,
            transport_key_id=transport_key_id,
            transport_key_needed=False
        )

    def test_new_secret_fails_with_invalid_transport_key_ref(self):
        resp, _ = create_secret(
            self.app,
            payload='superdupersecret',
            content_type='text/plain',
            transport_key_id="non_existing_transport_key_id",
            transport_key_needed="true",
            expect_errors=True
        )
        self.assertEqual(400, resp.status_int)

    def test_new_secret_w_unsupported_content_type_should_fail(self):
        resp, _ = create_secret(
            self.app,
            payload='something_here',
            content_type='bogus_content_type',
            expect_errors=True
        )
        self.assertEqual(400, resp.status_int)

    @utils.parameterized_dataset({
        'no_encoding': [None, 'application/octet-stream'],
        'bad_encoding': ['purple', 'application/octet-stream'],
        'no_content_type': ['base64', None]
    })
    def test_new_secret_fails_with_binary_payload_and(self, encoding=None,
                                                      content_type=None):
        resp, _ = create_secret(
            self.app,
            payload='lOtfqHaUUpe6NqLABgquYQ==',
            content_type=content_type,
            content_encoding=encoding,
            expect_errors=True
        )
        self.assertEqual(400, resp.status_int)

    def test_new_secret_fails_with_bad_payload(self):
        resp, _ = create_secret(
            self.app,
            payload='AAAAAAAAA',
            content_type='application/octet-stream',
            content_encoding='base64',
            expect_errors=True
        )
        self.assertEqual(400, resp.status_int)


class WhenGettingSecretsList(utils.BarbicanAPIBaseTestCase):

    def test_list_secrets_by_name(self):
        # Creating a secret to be retrieved later
        create_resp, _ = create_secret(
            self.app,
            name='secret mission'
        )

        self.assertEqual(201, create_resp.status_int)

        params = {'name': 'secret mission'}

        get_resp = self.app.get('/secrets/', params)

        self.assertEqual(200, get_resp.status_int)
        secret_list = get_resp.json.get('secrets')
        self.assertEqual('secret mission', secret_list[0].get('name'))

    def test_list_secrets(self):
        # Creating a secret to be retrieved later
        create_resp, _ = create_secret(
            self.app,
            name='James Bond'
        )

        self.assertEqual(201, create_resp.status_int)

        get_resp = self.app.get('/secrets/')

        self.assertEqual(200, get_resp.status_int,)
        self.assertIn('total', get_resp.json)
        secret_list = get_resp.json.get('secrets')
        self.assertGreater(len(secret_list), 0)

    def test_pagination_attributes(self):
        # Create a list of secrets greater than default limit (10)
        for _ in range(11):
            create_resp, _ = create_secret(self.app, name='Sterling Archer')
            self.assertEqual(201, create_resp.status_int)
        params = {'limit': '2', 'offset': '2'}

        get_resp = self.app.get('/secrets/', params)

        self.assertEqual(200, get_resp.status_int)
        self.assertIn('previous', get_resp.json)
        self.assertIn('next', get_resp.json)

        previous_ref = get_resp.json.get('previous')
        next_ref = get_resp.json.get('next')

        self.assertIn('offset=0', previous_ref)
        self.assertIn('offset=4', next_ref)

    def test_empty_list_of_secrets(self):
        params = {'name': 'Austin Powers'}

        get_resp = self.app.get('/secrets/', params)
        self.assertEqual(200, get_resp.status_int)

        secret_list = get_resp.json.get('secrets')
        self.assertEqual(0, len(secret_list))

        # These should never exist in this scenario
        self.assertNotIn('previous', get_resp.json)
        self.assertNotIn('next', get_resp.json)

    def test_bad_date_filter_results_in_400(self):
        params = {'expiration': 'bogus'}
        get_resp = self.app.get('/secrets/', params, expect_errors=True)
        self.assertEqual(400, get_resp.status_int)

    def test_bad_sorting_results_in_400(self):
        params = {'sort': 'bogus'}
        get_resp = self.app.get('/secrets/', params, expect_errors=True)
        self.assertEqual(400, get_resp.status_int)


class WhenGettingPuttingOrDeletingSecret(utils.BarbicanAPIBaseTestCase):

    def test_get_secret_as_plain(self):
        payload = 'this message will self destruct in 10 seconds'
        resp, secret_uuid = create_secret(
            self.app,
            payload=payload,
            content_type='text/plain'
        )

        self.assertEqual(201, resp.status_int)

        headers = {'Accept': 'text/plain'}
        get_resp = self.app.get(
            '/secrets/{0}'.format(secret_uuid), headers=headers
        )
        self.assertEqual(200, get_resp.status_int)
        self.assertEqual(payload, get_resp.body.decode('utf-8'))

    def test_get_secret_payload_with_pecan_default_accept_header(self):
        payload = 'a very interesting string'
        resp, secret_uuid = create_secret(
            self.app,
            payload=payload,
            content_type='text/plain'
        )

        self.assertEqual(201, resp.status_int)

        headers = {'Accept': '*/*'}
        get_resp = self.app.get(
            '/secrets/{0}/payload'.format(secret_uuid), headers=headers
        )
        self.assertEqual(200, get_resp.status_int)
        self.assertEqual(payload, get_resp.body.decode('utf-8'))

    def test_get_secret_payload_with_blank_accept_header(self):
        payload = 'a very interesting string'
        resp, secret_uuid = create_secret(
            self.app,
            payload=payload,
            content_type='text/plain'
        )

        self.assertEqual(201, resp.status_int)

        headers = {'Accept': ''}
        get_resp = self.app.get(
            '/secrets/{0}/payload'.format(secret_uuid), headers=headers
        )
        self.assertEqual(200, get_resp.status_int)
        self.assertEqual(payload, get_resp.body.decode('utf-8'))

    def test_get_secret_payload_with_no_accept_header(self):
        payload = 'a very interesting string'
        resp, secret_uuid = create_secret(
            self.app,
            payload=payload,
            content_type='text/plain'
        )

        self.assertEqual(201, resp.status_int)

        headers = {}
        get_resp = self.app.get(
            '/secrets/{0}/payload'.format(secret_uuid), headers=headers
        )
        self.assertEqual(200, get_resp.status_int)
        self.assertEqual(payload, get_resp.body.decode('utf-8'))

    def test_get_secret_is_decoded_for_binary(self):
        payload = 'a123'
        resp, secret_uuid = create_secret(
            self.app,
            payload=payload,
            content_type='application/octet-stream',
            content_encoding='base64'
        )
        headers = {
            'Accept': 'application/octet-stream',
        }
        get_resp = self.app.get(
            '/secrets/{0}'.format(secret_uuid), headers=headers
        )
        decoded = b'k]\xb7'

        self.assertEqual(decoded, get_resp.body)

    def test_returns_404_on_get_when_not_found(self):
        """Test with valid uuid which is not present in DB."""
        get_resp = self.app.get(
            '/secrets/{0}'.format(utils.generate_test_valid_uuid()),
            headers={'Accept': 'application/json'},
            expect_errors=True
        )
        self.assertEqual(404, get_resp.status_int)

    def test_returns_404_on_get_invalid_secret_id(self):
        """Test where uuid provided is not valid."""
        get_resp = self.app.get(
            '/secrets/98c876d9-aaac-44e4-8ea8-invalid-id',
            headers={'Accept': 'application/json'},
            expect_errors=True
        )
        self.assertEqual(404, get_resp.status_int)

    def test_returns_404_on_get_payload_when_no_payload(self):
        resp, secret_uuid = create_secret(self.app)
        headers = {
            'Accept': 'text/plain',
        }
        get_resp = self.app.get(
            '/secrets/{0}/payload'.format(secret_uuid),
            headers=headers,
            expect_errors=True
        )
        self.assertEqual(404, get_resp.status_int)

    def test_returns_404_on_get_with_bad_uuid(self):
        get_resp = self.app.get(
            '/secrets/98c876d9-aaac-44e4-8ea8-441932962b05X',
            headers={'Accept': 'application/json'},
            expect_errors=True
        )
        self.assertEqual(404, get_resp.status_int)

    def test_returns_406_with_get_bad_accept_header(self):
        resp, secret_uuid = create_secret(
            self.app,
            payload='blah',
            content_type='text/plain'
        )

        self.assertEqual(201, resp.status_int)

        get_resp = self.app.get(
            '/secrets/{0}'.format(secret_uuid),
            headers={'Accept': 'golden gun', 'Accept-Encoding': 'gzip'},
            expect_errors=True
        )

        self.assertEqual(406, get_resp.status_int)

    def test_put_plain_text_secret(self):
        resp, secret_uuid = create_secret(
            self.app
        )

        self.assertEqual(201, resp.status_int)

        message = 'Babou! Serpentine!'

        put_resp = self.app.put(
            '/secrets/{0}'.format(secret_uuid),
            message,
            headers={'Content-Type': 'text/plain'}
        )

        self.assertEqual(204, put_resp.status_int)

        get_resp = self.app.get(
            '/secrets/{0}'.format(secret_uuid),
            headers={'Accept': 'text/plain'}
        )

        self.assertEqual(200, get_resp.status_int)
        self.assertEqual(message, get_resp.body.decode('utf-8'))

    def test_put_binary_secret(self):
        resp, secret_uuid = create_secret(
            self.app
        )

        self.assertEqual(201, resp.status_int)

        binary_string = b'a binary string'
        put_resp = self.app.put(
            '/secrets/{0}'.format(secret_uuid),
            binary_string,
            headers={'Content-Type': 'application/octet-stream'}
        )

        self.assertEqual(204, put_resp.status_int)

        get_resp = self.app.get(
            '/secrets/{0}'.format(secret_uuid),
            headers={'Accept': 'application/octet-stream'}
        )

        self.assertEqual(200, get_resp.status_int)
        self.assertEqual(binary_string, get_resp.body)

    def test_put_base64_secret(self):
        resp, secret_uuid = create_secret(
            self.app
        )

        self.assertEqual(201, resp.status_int)

        payload = base64.b64encode(b'I had something for this')
        put_resp = self.app.put(
            '/secrets/{0}'.format(secret_uuid),
            payload,
            headers={
                'Content-Type': 'application/octet-stream',
                'Content-Encoding': 'base64'
            }
        )

        self.assertEqual(204, put_resp.status_int)

        get_resp = self.app.get(
            '/secrets/{0}'.format(secret_uuid),
            headers={
                'Accept': 'application/octet-stream',
                'Content-Encoding': 'base64'
            }
        )

        self.assertEqual(200, get_resp.status_int)
        self.assertEqual(base64.b64decode(payload), get_resp.body)

    def test_returns_400_with_put_unknown_encoding(self):
        resp, secret_uuid = create_secret(
            self.app
        )

        self.assertEqual(201, resp.status_int)

        payload = base64.b64encode(b'I had something for this')
        put_resp = self.app.put(
            '/secrets/{0}'.format(secret_uuid),
            payload,
            headers={
                'Accept': 'text/plain',
                'Content-Type': 'application/octet-stream',
                'Content-Encoding': 'unknownencoding'
            },
            expect_errors=True
        )

        self.assertEqual(400, put_resp.status_int)

    def test_returns_415_with_put_unsupported_media_type(self):
        resp, secret_uuid = create_secret(
            self.app
        )

        self.assertEqual(201, resp.status_int)

        put_resp = self.app.put(
            '/secrets/{0}'.format(secret_uuid),
            'rampage',
            headers={
                'Content-Type': 'application/json'
            },
            expect_errors=True
        )

        self.assertEqual(415, put_resp.status_int)

    def test_returns_415_with_put_no_media_type(self):
        resp, secret_uuid = create_secret(
            self.app
        )

        self.assertEqual(201, resp.status_int)

        put_resp = self.app.put(
            '/secrets/{0}'.format(secret_uuid),
            'rampage again',
            headers={
                'Content-Type': ''
            },
            expect_errors=True
        )

        self.assertEqual(415, put_resp.status_int)

    def test_returns_404_put_secret_not_found(self):
        put_resp = self.app.put(
            '/secrets/98c876d9-aaac-44e4-8ea8-441932962b05',
            'some text',
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(404, put_resp.status_int)

    def test_returns_409_put_to_existing_secret(self):
        resp, secret_uuid = create_secret(
            self.app,
            payload='blah',
            content_type='text/plain'
        )

        self.assertEqual(201, resp.status_int)

        put_resp = self.app.put(
            '/secrets/{0}'.format(secret_uuid),
            'do not want',
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(409, put_resp.status_int)

    def test_returns_400_put_no_payload(self):
        resp, secret_uuid = create_secret(
            self.app
        )

        self.assertEqual(201, resp.status_int)

        put_resp = self.app.put(
            '/secrets/{0}'.format(secret_uuid),
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(400, put_resp.status_int)

    def test_returns_400_put_with_empty_payload(self):
        resp, secret_uuid = create_secret(
            self.app
        )

        self.assertEqual(201, resp.status_int)

        put_resp = self.app.put(
            '/secrets/{0}'.format(secret_uuid),
            '',
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(400, put_resp.status_int)

    def test_returns_413_put_with_text_too_large(self):
        resp, secret_uuid = create_secret(
            self.app
        )

        self.assertEqual(201, resp.status_int)

        text_too_big = 'x' * 10050
        put_resp = self.app.put(
            '/secrets/{0}'.format(secret_uuid),
            text_too_big,
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(413, put_resp.status_int)

    def test_delete_secret(self):
        resp, secret_uuid = create_secret(
            self.app
        )

        self.assertEqual(201, resp.status_int)

        delete_resp = self.app.delete(
            '/secrets/{0}/'.format(secret_uuid)
        )

        self.assertEqual(204, delete_resp.status_int)

    def test_raise_404_for_delete_secret_not_found(self):
        delete_resp = self.app.delete(
            '/secrets/98c876d9-aaac-44e4-8ea8-441932962b05',
            expect_errors=True
        )

        self.assertEqual(404, delete_resp.status_int)
        self.assertEqual('application/json', delete_resp.content_type)

    def test_delete_with_json_accept_header(self):
        resp, secret_uuid = create_secret(
            self.app
        )

        self.assertEqual(201, resp.status_int)

        delete_resp = self.app.delete(
            '/secrets/{0}/'.format(secret_uuid),
            headers={'Accept': 'application/json'}
        )

        self.assertEqual(204, delete_resp.status_int)


@utils.parameterized_test_case
class WhenPerformingUnallowedOperations(utils.BarbicanAPIBaseTestCase):

    def test_returns_405_for_put_json_on_secrets(self):
        test_json = {
            'name': 'Barry',
            'algorithm': 'AES',
            'bit_length': 256,
            'mode': 'CBC'
        }
        resp = self.app.put_json(
            '/secrets/',
            test_json,
            expect_errors=True
        )

        self.assertEqual(405, resp.status_int)

    def test_returns_405_for_delete_on_secrets(self):
        resp = self.app.delete(
            '/secrets/',
            expect_errors=True
        )

        self.assertEqual(405, resp.status_int)

    def test_returns_405_for_get_payload(self):
        created_resp, secret_uuid = create_secret(
            self.app
        )
        resp = self.app.post(
            '/secrets/{0}/payload'.format(secret_uuid),
            'Do you want ants? This is how you get ants!',
            headers={'Content-Type': 'text/plain'},
            expect_errors=True
        )

        self.assertEqual(405, resp.status_int)

    @utils.parameterized_dataset({
        'delete': ['delete'],
        'put': ['put'],
        'post': ['post']
    })
    def test_returns_405_for_calling_secret_payload_uri_with(
            self, http_verb=None
    ):
        created_resp, secret_uuid = create_secret(
            self.app
        )

        self.assertEqual(201, created_resp.status_int)
        operation = getattr(self.app, http_verb)
        resp = operation(
            '/secrets/{0}/payload'.format(secret_uuid),
            'boop',
            expect_errors=True
        )

        self.assertEqual(405, resp.status_int)


class WhenValidatingDateFilters(utils.BarbicanAPIBaseTestCase):

    def setUp(self):
        super(WhenValidatingDateFilters, self).setUp()
        self.controller = secrets.SecretsController()

    def test_validates_plain_timestamp(self):
        date_filter = '2016-01-01T00:00:00'
        self.assertTrue(self.controller._is_valid_date_filter(date_filter))

    def test_validates_gt_and_lt_timestamps(self):
        date_filter = 'gt:2016-01-01T00:00:00,lt:2016-12-31T00:00:00'
        self.assertTrue(self.controller._is_valid_date_filter(date_filter))

    def test_validates_gte_and_lte_timestamps(self):
        date_filter = 'gte:2016-01-01T00:00:00,lte:2016-12-31T00:00:00'
        self.assertTrue(self.controller._is_valid_date_filter(date_filter))

    def test_validation_fails_with_two_plain_timestamps(self):
        date_filter = '2016-01-01T00:00:00,2016-01-02T00:00:00'
        self.assertFalse(self.controller._is_valid_date_filter(date_filter))

    def test_validation_fails_with_two_gt_timestamps(self):
        date_filter = 'gt:2016-01-01T00:00:00,gt:2016-01-02T00:00:00'
        self.assertFalse(self.controller._is_valid_date_filter(date_filter))

    def test_validation_fails_with_two_lt_timestamps(self):
        date_filter = 'lt:2016-01-01T00:00:00,lt:2016-01-02T00:00:00'
        self.assertFalse(self.controller._is_valid_date_filter(date_filter))

    def test_validation_fails_with_two_gte_timestamps(self):
        date_filter = 'gte:2016-01-01T00:00:00,gte:2016-01-02T00:00:00'
        self.assertFalse(self.controller._is_valid_date_filter(date_filter))

    def test_validation_fails_with_two_lte_timestamps(self):
        date_filter = 'lte:2016-01-01T00:00:00,lte:2016-01-02T00:00:00'
        self.assertFalse(self.controller._is_valid_date_filter(date_filter))

    def test_validation_fails_with_plain_and_gte_timestamps(self):
        date_filter = '2016-01-01T00:00:00,gte:2016-01-02T00:00:00'
        self.assertFalse(self.controller._is_valid_date_filter(date_filter))

        date_filter = 'gte:2016-01-01T00:00:00,2016-01-02T00:00:00'
        self.assertFalse(self.controller._is_valid_date_filter(date_filter))

    def test_validation_fails_with_plain_and_lte_timestamps(self):
        date_filter = '2016-01-01T00:00:00,lte:2016-01-02T00:00:00'
        self.assertFalse(self.controller._is_valid_date_filter(date_filter))

        date_filter = 'lte:2016-01-01T00:00:00,2016-01-02T00:00:00'
        self.assertFalse(self.controller._is_valid_date_filter(date_filter))

    def test_validation_fails_with_gt_and_gte_timestamps(self):
        date_filter = 'gt:2016-01-01T00:00:00,gte:2016-01-02T00:00:00'
        self.assertFalse(self.controller._is_valid_date_filter(date_filter))

    def test_validation_fails_with_lt_and_lte_timestamps(self):
        date_filter = 'lt:2016-01-01T00:00:00,lte:2016-01-02T00:00:00'
        self.assertFalse(self.controller._is_valid_date_filter(date_filter))

    def test_validation_fails_with_bogus_timestamp(self):
        date_filter = 'bogus'
        self.assertFalse(self.controller._is_valid_date_filter(date_filter))


class WhenValidatingSortFilters(utils.BarbicanAPIBaseTestCase):

    def setUp(self):
        super(WhenValidatingSortFilters, self).setUp()
        self.controller = secrets.SecretsController()

    def test_validates_name_sorting(self):
        sorting = 'name'
        self.assertTrue(self.controller._is_valid_sorting(sorting))

    def test_validation_fails_for_bogus_attribute(self):
        sorting = 'bogus'
        self.assertFalse(self.controller._is_valid_sorting(sorting))

    def test_validation_fails_for_duplicate_keys(self):
        sorting = 'name,name:asc'
        self.assertFalse(self.controller._is_valid_sorting(sorting))

    def test_validation_fails_for_too_many_colons(self):
        sorting = 'name:asc:foo'
        self.assertFalse(self.controller._is_valid_sorting(sorting))


# ----------------------- Helper Functions ---------------------------
def create_secret(app, name=None, algorithm=None, bit_length=None, mode=None,
                  expiration=None, payload=None, content_type=None,
                  content_encoding=None, transport_key_id=None,
                  transport_key_needed=None, expect_errors=False):
    # TODO(chellygel): Once test resources is split out, refactor this
    # and similar functions into a generalized helper module and reduce
    # duplication.
    request = {
        'name': name,
        'algorithm': algorithm,
        'bit_length': bit_length,
        'mode': mode,
        'expiration': expiration,
        'payload': payload,
        'payload_content_type': content_type,
        'payload_content_encoding': content_encoding,
        'transport_key_id': transport_key_id,
        'transport_key_needed': transport_key_needed
    }
    cleaned_request = {key: val for key, val in request.items()
                       if val is not None}

    resp = app.post_json(
        '/secrets/',
        cleaned_request,
        expect_errors=expect_errors
    )

    created_uuid = None
    if resp.status_int == 201:
        secret_ref = resp.json.get('secret_ref', '')
        _, created_uuid = os.path.split(secret_ref)

    return resp, created_uuid



"""
Copyright 2014-2015 Rackspace

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import logging
from oslo_serialization import jsonutils as json

LOG = logging.getLogger(__name__)


class BaseModel(object):
    """Base class for models.

    To allow simple (de)serialization we will use __dict__ to create
    """
    def override_values(self, **kwargs):
        for attr_name, attr_value in kwargs.items():
            if hasattr(self, attr_name):
                setattr(self, attr_name, attr_value)

    def obj_to_json(self):
        """Convert this object to a JSON string.

        :return: A string of JSON containing the fields in this object
        """
        return json.dump_as_bytes(self.obj_to_dict())

    def obj_to_dict(self):
        """Create a dict of the values for this model object.

        If there are  fields that are not set in this object then those
        will NOT have entries in the returned dict.

        :return: A dict representing this model
        """
        the_dict = self.__dict__
        retval = self._remove_empty_fields_from_dict(the_dict)
        return retval

    def _remove_empty_fields_from_dict(self, dictionary):
        """Remove k,v pairs with empty values from a dictionary.

        :param dictionary: a dictionary of stuff
        :return: the same dictionary where all k,v pairs with empty values
        have been removed.
        """

        # NOTE(jaosorior): deleting a key from the incoming dictionary actually
        # affects the model object. So we do a copy to avoid this.
        resulting_dict = dictionary.copy()

        # Dumping the keys to a list as we'll be changing the dict size
        empty_keys = [k for k, v in dictionary.items() if v is None]
        for k in empty_keys:
            del resulting_dict[k]
        return resulting_dict

    @classmethod
    def json_to_obj(cls, serialized_str):
        """Create a model from a JSON string.

        :param serialized_str: the JSON string
        :return: a secret object
        """
        try:
            json_dict = json.loads(serialized_str)
            return cls.dict_to_obj(json_dict)
        except TypeError as e:
            LOG.error('Couldn\'t deserialize input: %s\n Because: %s',
                      serialized_str, e)

    @classmethod
    def dict_to_obj(cls, input_dict):
        """Create an object from a dict.

        :param input_dict: A dict of fields.
        :return: a model object build from the passed in dict.
        """
        return cls(**input_dict)



# Copyright (c) 2013-2014 Rackspace, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ldap3.utils.dn import parse_dn
from OpenSSL import crypto

from barbican.common import exception as excep
from barbican.common import hrefs
from barbican.common import resources as res
import barbican.common.utils as utils
from barbican.model import models
from barbican.model import repositories as repos
from barbican.plugin.interface import certificate_manager as cert
from barbican.plugin import resources as plugin
from barbican.tasks import common

LOG = utils.getLogger(__name__)

# Order sub-status definitions
ORDER_STATUS_REQUEST_PENDING = models.OrderStatus(
    "cert_request_pending",
    "Request has been submitted to the CA.  "
    "Waiting for certificate to be generated"
)

ORDER_STATUS_CERT_GENERATED = models.OrderStatus(
    "cert_generated",
    "Certificate has been generated"
)

ORDER_STATUS_DATA_INVALID = models.OrderStatus(
    "cert_data_invalid",
    "CA rejected request data as invalid"
)

ORDER_STATUS_CA_UNAVAIL_FOR_ISSUE = models.OrderStatus(
    "cert_ca_unavail_for_issue",
    "Unable to submit certificate request.  CA unavailable"
)

ORDER_STATUS_INVALID_OPERATION = models.OrderStatus(
    "cert_invalid_operation",
    "CA returned invalid operation"
)

ORDER_STATUS_INTERNAL_ERROR = models.OrderStatus(
    "cert_internal_error",
    "Internal error during certificate operations"
)

ORDER_STATUS_CA_UNAVAIL_FOR_CHECK = models.OrderStatus(
    "cert_ca_unavail_for_status_check",
    "Unable to get certificate request status.  CA unavailable."
)


def refresh_certificate_resources():
    # Before CA operations can be performed, the CA table must be populated
    cert.CertificatePluginManager().refresh_ca_table()


def issue_certificate_request(order_model, project_model, result_follow_on):
    """Create the initial order with CA.

    Note that this method may be called more than once if retries are
    required. Barbican metadata is used to store intermediate information,
    including selected plugins by name, to support such retries.

    :param: order_model - order associated with this cert request
    :param: project_model - project associated with this request
    :param: result_follow_on - A :class:`FollowOnProcessingStatusDTO` instance
        instantiated by the client that this function may optionally update
        with information on how to process this task into the future.
    :returns: container_model - container with the relevant cert if
        the request has been completed.  None otherwise
    """
    plugin_meta = _get_plugin_meta(order_model)
    barbican_meta = _get_barbican_meta(order_model)

    # TODO(john-wood-w) We need to de-conflict barbican_meta (stored with order
    # and not shown to plugins) with barbican_meta_dto (shared with plugins).
    # As a minimum we should change the name of the DTO to something like
    # 'extended_meta_dto' or some such.
    barbican_meta_for_plugins_dto = cert.BarbicanMetaDTO()

    # refresh the CA table.  This is mostly a no-op unless the entries
    # for a plugin are expired.
    cert.CertificatePluginManager().refresh_ca_table()

    cert_plugin = _get_cert_plugin(barbican_meta,
                                   barbican_meta_for_plugins_dto,
                                   order_model, project_model)
    barbican_meta['plugin_name'] = utils.generate_fullname_for(cert_plugin)

    # Generate CSR if needed.
    request_type = order_model.meta.get(cert.REQUEST_TYPE)
    if request_type == cert.CertificateRequestType.STORED_KEY_REQUEST:
        csr = barbican_meta.get('generated_csr')
        if csr is None:
            # TODO(alee) Fix this to be a non-project specific call once
            # the ACL patches go in.
            csr = _generate_csr_from_private_key(order_model, project_model)
            barbican_meta['generated_csr'] = csr
        barbican_meta_for_plugins_dto.generated_csr = csr

    result = cert_plugin.issue_certificate_request(
        order_model.id, order_model.meta,
        plugin_meta, barbican_meta_for_plugins_dto)

    # Save plugin and barbican metadata for this order.
    _save_plugin_metadata(order_model, plugin_meta)
    _save_barbican_metadata(order_model, barbican_meta)

    # Handle result
    return _handle_task_result(
        result, result_follow_on, order_model, project_model, request_type,
        unavailable_status=ORDER_STATUS_CA_UNAVAIL_FOR_ISSUE)


def _get_cert_plugin(barbican_meta, barbican_meta_for_plugins_dto,
                     order_model, project_model):
    cert_plugin_name = barbican_meta.get('plugin_name')
    if cert_plugin_name:
        return cert.CertificatePluginManager().get_plugin_by_name(
            cert_plugin_name)
    ca_id = _get_ca_id(order_model.meta, project_model.id)
    if ca_id:
        ca = repos.get_ca_repository().get(ca_id)
        barbican_meta_for_plugins_dto.plugin_ca_id = ca.plugin_ca_id
        return cert.CertificatePluginManager().get_plugin_by_name(
            ca.plugin_name)
    else:
        return cert.CertificatePluginManager().get_plugin(order_model.meta)


def check_certificate_request(order_model, project_model, result_follow_on):
    """Check the status of a certificate request with the CA.

    Note that this method may be called more than once if retries are
    required. Barbican metadata is used to store intermediate information,
    including selected plugins by name, to support such retries.

    :param: order_model - order associated with this cert request
    :param: project_model - project associated with this request
    :param: result_follow_on - A :class:`FollowOnProcessingStatusDTO` instance
        instantiated by the client that this function may optionally update
        with information on how to process this task into the future.
    :returns: container_model - container with the relevant cert if the
        request has been completed.  None otherwise.
    """
    plugin_meta = _get_plugin_meta(order_model)
    barbican_meta = _get_barbican_meta(order_model)

    # TODO(john-wood-w) See note above about DTO's name.
    barbican_meta_for_plugins_dto = cert.BarbicanMetaDTO()

    cert_plugin = cert.CertificatePluginManager().get_plugin_by_name(
        barbican_meta.get('plugin_name'))

    result = cert_plugin.check_certificate_status(
        order_model.id, order_model.meta,
        plugin_meta, barbican_meta_for_plugins_dto)

    # Save plugin order plugin state
    _save_plugin_metadata(order_model, plugin_meta)

    request_type = order_model.meta.get(cert.REQUEST_TYPE)
    return _handle_task_result(
        result, result_follow_on, order_model, project_model, request_type,
        unavailable_status=ORDER_STATUS_CA_UNAVAIL_FOR_CHECK)


def create_subordinate_ca(project_model, name, description, subject_dn,
                          parent_ca_ref, creator_id):
    """Create a subordinate CA

    :param name - name of the subordinate CA
    :param: description - description of the subordinate CA
    :param: subject_dn - subject DN of the subordinate CA
    :param: parent_ca_ref - Barbican URL reference to the parent CA
    :param: creator_id - id for creator of the subordinate CA
    :return: :class models.CertificateAuthority model object for new sub CA
    """
    # check that the parent ref exists and is accessible
    parent_ca_id = hrefs.get_ca_id_from_ref(parent_ca_ref)
    ca_repo = repos.get_ca_repository()
    parent_ca = ca_repo.get(entity_id=parent_ca_id, suppress_exception=True)
    if not parent_ca:
        raise excep.InvalidParentCA(parent_ca_ref=parent_ca_ref)

    # Parent CA must be a base CA or a subCA owned by this project
    if (parent_ca.project_id is not None and
            parent_ca.project_id != project_model.id):
        raise excep.UnauthorizedSubCA()

    # get the parent plugin, raises CertPluginNotFound if missing
    cert_plugin = cert.CertificatePluginManager().get_plugin_by_name(
        parent_ca.plugin_name)

    # confirm that the plugin supports creating subordinate CAs
    if not cert_plugin.supports_create_ca():
        raise excep.SubCAsNotSupported()

    # make call to create the subordinate ca
    create_ca_dto = cert.CACreateDTO(
        name=name,
        description=description,
        subject_dn=subject_dn,
        parent_ca_id=parent_ca.plugin_ca_id)

    new_ca_dict = cert_plugin.create_ca(create_ca_dto)
    if not new_ca_dict:
        raise excep.SubCANotCreated(name=name)

    # create and store the subordinate CA as a new certificate authority object
    new_ca_dict['plugin_name'] = parent_ca.plugin_name
    new_ca_dict['creator_id'] = creator_id
    new_ca_dict['project_id'] = project_model.id
    new_ca = models.CertificateAuthority(new_ca_dict)
    ca_repo.create_from(new_ca)

    return new_ca


def delete_subordinate_ca(external_project_id, ca):
    """Deletes a subordinate CA and any related artifacts

    :param external_project_id: external project ID
    :param ca: class:`models.CertificateAuthority` to be deleted
    :return: None
     """
    # TODO(alee) See if the checks below can be moved to the RBAC code

    # Check that this CA is a subCA
    if ca.project_id is None:
        raise excep.CannotDeleteBaseCA()

    # Check that the user's project owns this subCA
    project = res.get_or_create_project(external_project_id)
    if ca.project_id != project.id:
        raise excep.UnauthorizedSubCA()

    project_ca_repo = repos.get_project_ca_repository()
    (project_cas, _, _, _) = project_ca_repo.get_by_create_date(
        project_id=project.id, ca_id=ca.id,
        suppress_exception=True)

    preferred_ca_repo = repos.get_preferred_ca_repository()
    (preferred_cas, _, _, _) = preferred_ca_repo.get_by_create_date(
        project_id=project.id, ca_id=ca.id, suppress_exception=True)

    # Can not delete a project preferred CA, if other project CAs exist. One
    # of those needs to be designated as the preferred CA first.
    if project_cas and preferred_cas and not is_last_project_ca(project.id):
        raise excep.CannotDeletePreferredCA()

    # Remove the CA as preferred
    if preferred_cas:
        preferred_ca_repo.delete_entity_by_id(preferred_cas[0].id,
                                              external_project_id)
    # Remove the CA from project list
    if project_cas:
        project_ca_repo.delete_entity_by_id(project_cas[0].id,
                                            external_project_id)

    # Delete the CA entry from plugin
    cert_plugin = cert.CertificatePluginManager().get_plugin_by_name(
        ca.plugin_name)
    cert_plugin.delete_ca(ca.plugin_ca_id)

    # Finally, delete the CA entity from the CA repository
    ca_repo = repos.get_ca_repository()
    ca_repo.delete_entity_by_id(
        entity_id=ca.id,
        external_project_id=external_project_id)


def is_last_project_ca(project_id):
    """Returns True iff project has exactly one project CA

    :param project_id: internal project ID
    :return: Boolean
     """
    project_ca_repo = repos.get_project_ca_repository()
    _, _, _, total = project_ca_repo.get_by_create_date(
        project_id=project_id,
        suppress_exception=True
    )
    return total == 1


def _handle_task_result(result, result_follow_on, order_model,
                        project_model, request_type, unavailable_status):
    if cert.CertificateStatus.WAITING_FOR_CA == result.status:
        _update_result_follow_on(
            result_follow_on,
            order_status=ORDER_STATUS_REQUEST_PENDING,
            retry_task=common.RetryTasks.INVOKE_CERT_STATUS_CHECK_TASK,
            retry_msec=result.retry_msec)
    elif cert.CertificateStatus.CERTIFICATE_GENERATED == result.status:
        _update_result_follow_on(
            result_follow_on,
            order_status=ORDER_STATUS_CERT_GENERATED)
        container_model = _save_secrets(result, project_model, request_type,
                                        order_model)
        return container_model
    elif cert.CertificateStatus.CLIENT_DATA_ISSUE_SEEN == result.status:
        raise cert.CertificateStatusClientDataIssue(result.status_message)
    elif cert.CertificateStatus.CA_UNAVAILABLE_FOR_REQUEST == result.status:
        _update_result_follow_on(
            result_follow_on,
            order_status=unavailable_status,
            retry_task=common.RetryTasks.INVOKE_SAME_TASK,
            retry_msec=cert.ERROR_RETRY_MSEC)
        _notify_ca_unavailable(order_model, result)
    elif cert.CertificateStatus.INVALID_OPERATION == result.status:
        raise cert.CertificateStatusInvalidOperation(result.status_message)
    else:
        raise cert.CertificateStatusNotSupported(result.status)

    return None


def _add_private_key_to_generated_cert_container(container_id, order_model,
                                                 project_model):
    keypair_container_id, keypair_container = _get_container_from_order_meta(
        order_model, project_model)
    private_key_id = None

    for cs in keypair_container.container_secrets:
        if cs.name == 'private_key':
            private_key_id = cs.secret_id

    new_consec_assoc = models.ContainerSecret()
    new_consec_assoc.name = 'private_key'
    new_consec_assoc.container_id = container_id
    new_consec_assoc.secret_id = private_key_id
    container_secret_repo = repos.get_container_secret_repository()
    container_secret_repo.create_from(new_consec_assoc)


def modify_certificate_request(order_model, updated_meta):
    """Update the order with CA."""
    # TODO(chellygel): Add the modify certificate request logic.
    LOG.debug('in modify_certificate_request')
    raise NotImplementedError  # pragma: no cover


def get_global_preferred_ca():
    project = res.get_or_create_global_preferred_project()
    preferred_ca_repository = repos.get_preferred_ca_repository()
    cas = preferred_ca_repository.get_project_entities(project.id)
    if not cas:
        return None
    else:
        return cas[0]


def get_project_preferred_ca_id(project_id):
    """Compute the preferred CA ID for a project

    First priority: a preferred CA is defined for the project
    Second priority: a preferred CA is defined globally
    Else: None
    """
    preferred_ca_repository = repos.get_preferred_ca_repository()
    cas, offset, limit, total = preferred_ca_repository.get_by_create_date(
        project_id=project_id, suppress_exception=True)
    if total > 0:
        return cas[0].ca_id
    global_ca = get_global_preferred_ca()
    if global_ca:
        return global_ca.ca_id


def _get_ca_id(order_meta, project_id):
    ca_id = order_meta.get(cert.CA_ID)
    if ca_id:
        return ca_id

    return get_project_preferred_ca_id(project_id)


def _update_result_follow_on(
        result_follow_on,
        order_status=None,
        retry_task=common.RetryTasks.NO_ACTION_REQUIRED,
        retry_msec=common.RETRY_MSEC_DEFAULT):
    if order_status:
        result_follow_on.status = order_status.id
        result_follow_on.status_message = order_status.message
    result_follow_on.retry_task = retry_task
    if retry_msec and retry_msec >= 0:
        result_follow_on.retry_msec = retry_msec


def _get_plugin_meta(order_model):
    if order_model:
        order_plugin_meta_repo = repos.get_order_plugin_meta_repository()
        return order_plugin_meta_repo.get_metadata_for_order(order_model.id)
    else:
        return {}


def _get_barbican_meta(order_model):
    if order_model:
        order_barbican_meta_repo = repos.get_order_barbican_meta_repository()
        return order_barbican_meta_repo.get_metadata_for_order(order_model.id)
    else:
        return {}


def _generate_csr_from_private_key(order_model, project_model):
    """Generate a CSR from the private key.

    :param: order_model - order for the request
    :param: project_model - project for this request
    :return: CSR (certificate signing request) in PEM format
    :raise: :class:`StoredKeyPrivateKeyNotFound` if private key not found
            :class:`StoredKeyContainerNotFound` if container not found
    """
    container_id, container = _get_container_from_order_meta(order_model,
                                                             project_model)

    if not container:
        raise excep.StoredKeyContainerNotFound(container_id)

    passphrase = None
    private_key = None

    for cs in container.container_secrets:
        secret_repo = repos.get_secret_repository()
        if cs.name == 'private_key':
            private_key_model = secret_repo.get(
                cs.secret_id,
                project_model.external_id)
            private_key = plugin.get_secret(
                'application/pkcs8',
                private_key_model,
                project_model)
        elif cs.name == 'private_key_passphrase':
            passphrase_model = secret_repo.get(
                cs.secret_id,
                project_model.external_id)
            passphrase = plugin.get_secret(
                'text/plain;charset=utf-8',
                passphrase_model,
                project_model)
            passphrase = str(passphrase)

    if not private_key:
        raise excep.StoredKeyPrivateKeyNotFound(container.id)

    if passphrase is None:
        pkey = crypto.load_privatekey(
            crypto.FILETYPE_PEM,
            private_key
        )
    else:
        pkey = crypto.load_privatekey(
            crypto.FILETYPE_PEM,
            private_key,
            passphrase.encode('utf-8')
        )

    subject_name = order_model.meta.get('subject_dn')
    subject_name_dns = parse_dn(subject_name)
    extensions = order_model.meta.get('extensions', None)

    req = crypto.X509Req()
    subj = req.get_subject()

    # Note: must iterate over the DNs in reverse order, or the resulting
    # subject name will be reversed.
    for ava in reversed(subject_name_dns):
        key, val, extra = ava
        setattr(subj, key.upper(), val)
    req.set_pubkey(pkey)
    if extensions:
        # TODO(alee-3) We need code here to parse the encoded extensions and
        # convert them into X509Extension objects.  This code will also be
        # used in the validation code.  Commenting out for now till we figure
        # out how to do this.
        # req.add_extensions(extensions)
        pass
    req.sign(pkey, 'sha256')

    csr = crypto.dump_certificate_request(crypto.FILETYPE_PEM, req)
    return csr


def _get_container_from_order_meta(order_model, project_model):
    container_ref = order_model.meta.get('container_ref')

    # extract container_id as the last part of the URL
    container_id = hrefs.get_container_id_from_ref(container_ref)

    container_repo = repos.get_container_repository()
    container = container_repo.get(container_id,
                                   project_model.external_id,
                                   suppress_exception=True)
    return container_id, container


def _notify_ca_unavailable(order_model, result):
    """Notify observer(s) that the CA was unavailable at this time."""
    cert.get_event_plugin_manager().notify_ca_is_unavailable(
        order_model.project_id,
        hrefs.convert_order_to_href(order_model.id),
        result.status_message,
        result.retry_msec)


def _save_plugin_metadata(order_model, plugin_meta):
    """Add plugin metadata to an order."""

    if not isinstance(plugin_meta, dict):
        plugin_meta = {}

    order_plugin_meta_repo = repos.get_order_plugin_meta_repository()
    order_plugin_meta_repo.save(plugin_meta, order_model)


def _save_barbican_metadata(order_model, barbican_meta):
    """Add barbican metadata to an order."""

    if not isinstance(barbican_meta, dict):
        barbican_meta = {}

    order_barbican_meta_repo = repos.get_order_barbican_meta_repository()
    order_barbican_meta_repo.save(barbican_meta, order_model)


def _save_secrets(result, project_model, request_type, order_model):
    cert_secret_model, transport_key_model = plugin.store_secret(
        unencrypted_raw=result.certificate,
        content_type_raw='application/octet-stream',
        content_encoding='base64',
        secret_model=models.Secret(),
        project_model=project_model)

    # save the certificate chain as a secret.
    if result.intermediates:
        intermediates_secret_model, transport_key_model = plugin.store_secret(
            unencrypted_raw=result.intermediates,
            content_type_raw='application/octet-stream',
            content_encoding='base64',
            secret_model=models.Secret(),
            project_model=project_model
        )
    else:
        intermediates_secret_model = None

    container_model = models.Container()
    container_model.type = "certificate"
    container_model.status = models.States.ACTIVE
    container_model.project_id = project_model.id
    container_repo = repos.get_container_repository()
    container_repo.create_from(container_model)

    # create container_secret for certificate
    new_consec_assoc = models.ContainerSecret()
    new_consec_assoc.name = 'certificate'
    new_consec_assoc.container_id = container_model.id
    new_consec_assoc.secret_id = cert_secret_model.id
    container_secret_repo = repos.get_container_secret_repository()
    container_secret_repo.create_from(new_consec_assoc)

    if intermediates_secret_model:
        # create container_secret for intermediate certs
        new_consec_assoc = models.ContainerSecret()
        new_consec_assoc.name = 'intermediates'
        new_consec_assoc.container_id = container_model.id
        new_consec_assoc.secret_id = intermediates_secret_model.id
        container_secret_repo.create_from(new_consec_assoc)

    if request_type == cert.CertificateRequestType.STORED_KEY_REQUEST:
        _add_private_key_to_generated_cert_container(container_model.id,
                                                     order_model,
                                                     project_model)

    return container_model



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

import pecan

from barbican import api
from barbican.api import controllers
from barbican.api.controllers import acls
from barbican.api.controllers import consumers
from barbican.common import exception
from barbican.common import hrefs
from barbican.common import quota
from barbican.common import resources as res
from barbican.common import utils
from barbican.common import validators
from barbican import i18n as u
from barbican.model import models
from barbican.model import repositories as repo

LOG = utils.getLogger(__name__)

CONTAINER_GET = 'container:get'


def container_not_found():
    """Throw exception indicating container not found."""
    pecan.abort(404, u._('Secrets container not found.'))


def invalid_container_id():
    """Throw exception indicating container id is invalid."""
    pecan.abort(404, u._('Not Found. Provided container id is invalid.'))


class ContainerController(controllers.ACLMixin):
    """Handles Container entity retrieval and deletion requests."""

    def __init__(self, container):
        self.container = container
        self.container_id = container.id
        self.consumer_repo = repo.get_container_consumer_repository()
        self.container_repo = repo.get_container_repository()
        self.validator = validators.ContainerValidator()
        self.consumers = consumers.ContainerConsumersController(
            self.container_id)
        self.acl = acls.ContainerACLsController(self.container)

    def get_acl_tuple(self, req, **kwargs):
        d = self.get_acl_dict_for_user(req, self.container.container_acls)
        d['project_id'] = self.container.project.external_id
        d['creator_id'] = self.container.creator_id
        return 'container', d

    @pecan.expose(generic=True, template='json')
    def index(self, **kwargs):
        pecan.abort(405)  # HTTP 405 Method Not Allowed as default

    @index.when(method='GET', template='json')
    @controllers.handle_exceptions(u._('Container retrieval'))
    @controllers.enforce_rbac(CONTAINER_GET)
    def on_get(self, external_project_id):
        dict_fields = self.container.to_dict_fields()

        for secret_ref in dict_fields['secret_refs']:
            hrefs.convert_to_hrefs(secret_ref)

        LOG.info('Retrieved container for project: %s',
                 external_project_id)
        return hrefs.convert_to_hrefs(
            hrefs.convert_to_hrefs(dict_fields)
        )

    @index.when(method='DELETE')
    @utils.allow_all_content_types
    @controllers.handle_exceptions(u._('Container deletion'))
    @controllers.enforce_rbac('container:delete')
    def on_delete(self, external_project_id, **kwargs):
        container_consumers = self.consumer_repo.get_by_container_id(
            self.container_id,
            suppress_exception=True
        )
        try:
            self.container_repo.delete_entity_by_id(
                entity_id=self.container_id,
                external_project_id=external_project_id
            )
        except exception.NotFound:
            LOG.exception('Problem deleting container')
            container_not_found()

        LOG.info('Deleted container for project: %s',
                 external_project_id)

        for consumer in container_consumers[0]:
            try:
                self.consumer_repo.delete_entity_by_id(
                    consumer.id, external_project_id)
            except exception.NotFound:  # nosec
                pass


class ContainersController(controllers.ACLMixin):
    """Handles Container creation requests."""

    def __init__(self):
        self.consumer_repo = repo.get_container_consumer_repository()
        self.container_repo = repo.get_container_repository()
        self.secret_repo = repo.get_secret_repository()
        self.validator = validators.ContainerValidator()
        self.quota_enforcer = quota.QuotaEnforcer('containers',
                                                  self.container_repo)

    @pecan.expose()
    def _lookup(self, container_id, *remainder):
        if not utils.validate_id_is_uuid(container_id):
            invalid_container_id()
        container = self.container_repo.get_container_by_id(
            entity_id=container_id, suppress_exception=True)
        if not container:
            container_not_found()

        if len(remainder) > 0 and remainder[0] == 'secrets':
            return ContainersSecretsController(container), ()

        return ContainerController(container), remainder

    @pecan.expose(generic=True, template='json')
    def index(self, **kwargs):
        pecan.abort(405)  # HTTP 405 Method Not Allowed as default

    @index.when(method='GET', template='json')
    @controllers.handle_exceptions(u._('Containers(s) retrieval'))
    @controllers.enforce_rbac('containers:get')
    def on_get(self, project_id, **kw):
        LOG.debug('Start containers on_get for project-ID %s:', project_id)

        result = self.container_repo.get_by_create_date(
            project_id,
            offset_arg=kw.get('offset', 0),
            limit_arg=kw.get('limit', None),
            name_arg=kw.get('name', None),
            type_arg=kw.get('type', None),
            suppress_exception=True
        )

        containers, offset, limit, total = result

        if not containers:
            resp_ctrs_overall = {'containers': [], 'total': total}
        else:
            resp_ctrs = [
                hrefs.convert_to_hrefs(c.to_dict_fields())
                for c in containers
            ]

            for ctr in resp_ctrs:
                for secret_ref in ctr.get('secret_refs', []):
                    hrefs.convert_to_hrefs(secret_ref)

            resp_ctrs_overall = hrefs.add_nav_hrefs(
                'containers',
                offset,
                limit,
                total,
                {'containers': resp_ctrs}
            )
            resp_ctrs_overall.update({'total': total})

        LOG.info('Retrieved container list for project: %s', project_id)
        return resp_ctrs_overall

    @index.when(method='POST', template='json')
    @controllers.handle_exceptions(u._('Container creation'))
    @controllers.enforce_rbac('containers:post')
    @controllers.enforce_content_types(['application/json'])
    def on_post(self, external_project_id, **kwargs):

        project = res.get_or_create_project(external_project_id)

        data = api.load_body(pecan.request, validator=self.validator)
        ctxt = controllers._get_barbican_context(pecan.request)
        if ctxt:  # in authenticated pipleline case, always use auth token user
            data['creator_id'] = ctxt.user

        self.quota_enforcer.enforce(project)

        LOG.debug('Start on_post...%s', data)

        new_container = models.Container(data)
        new_container.project_id = project.id

        # TODO(hgedikli): performance optimizations
        for secret_ref in new_container.container_secrets:
            secret = self.secret_repo.get(
                entity_id=secret_ref.secret_id,
                external_project_id=external_project_id,
                suppress_exception=True)
            if not secret:
                # This only partially localizes the error message and
                # doesn't localize secret_ref.name.
                pecan.abort(
                    404,
                    u._("Secret provided for '{secret_name}' doesn't "
                        "exist.").format(secret_name=secret_ref.name)
                )

        self.container_repo.create_from(new_container)

        url = hrefs.convert_container_to_href(new_container.id)

        pecan.response.status = 201
        pecan.response.headers['Location'] = url
        LOG.info('Created a container for project: %s',
                 external_project_id)

        return {'container_ref': url}


class ContainersSecretsController(controllers.ACLMixin):
    """Handles ContainerSecret creation and deletion requests."""

    def __init__(self, container):
        LOG.debug('=== Creating ContainerSecretsController ===')
        self.container = container
        self.container_secret_repo = repo.get_container_secret_repository()
        self.secret_repo = repo.get_secret_repository()
        self.validator = validators.ContainerSecretValidator()

    @pecan.expose(generic=True)
    def index(self, **kwargs):
        pecan.abort(405)  # HTTP 405 Method Not Allowed as default

    @index.when(method='POST', template='json')
    @controllers.handle_exceptions(u._('Container Secret creation'))
    @controllers.enforce_rbac('container_secret:post')
    @controllers.enforce_content_types(['application/json'])
    def on_post(self, external_project_id, **kwargs):
        """Handles adding an existing secret to an existing container."""

        if self.container.type != 'generic':
            pecan.abort(400, u._("Only 'generic' containers can be modified."))

        data = api.load_body(pecan.request, validator=self.validator)

        name = data.get('name')
        secret_ref = data.get('secret_ref')
        secret_id = hrefs.get_secret_id_from_ref(secret_ref)

        secret = self.secret_repo.get(
            entity_id=secret_id,
            external_project_id=external_project_id,
            suppress_exception=True)
        if not secret:
            pecan.abort(404, u._("Secret provided doesn't exist."))

        found_container_secrets = list(
            filter(lambda cs: cs.secret_id == secret_id and cs.name == name,
                   self.container.container_secrets)
        )

        if found_container_secrets:
            pecan.abort(409, u._('Conflict. A secret with that name and ID is '
                                 'already stored in this container. The same '
                                 'secret can exist in a container as long as '
                                 'the name is unique.'))

        LOG.debug('Start container secret on_post...%s', secret_ref)
        new_container_secret = models.ContainerSecret()
        new_container_secret.container_id = self.container.id
        new_container_secret.name = name
        new_container_secret.secret_id = secret_id
        self.container_secret_repo.save(new_container_secret)

        url = hrefs.convert_container_to_href(self.container.id)
        LOG.debug('URI to container is %s', url)

        pecan.response.status = 201
        pecan.response.headers['Location'] = url
        LOG.info('Created a container secret for project: %s',
                 external_project_id)

        return {'container_ref': url}

    @index.when(method='DELETE')
    @utils.allow_all_content_types
    @controllers.handle_exceptions(u._('Container Secret deletion'))
    @controllers.enforce_rbac('container_secret:delete')
    def on_delete(self, external_project_id, **kwargs):
        """Handles removing a secret reference from an existing container."""

        data = api.load_body(pecan.request, validator=self.validator)

        name = data.get('name')
        secret_ref = data.get('secret_ref')
        secret_id = hrefs.get_secret_id_from_ref(secret_ref)

        secret = self.secret_repo.get(
            entity_id=secret_id,
            external_project_id=external_project_id,
            suppress_exception=True)
        if not secret:
            pecan.abort(404, u._("Secret '{secret_name}' with reference "
                                 "'{secret_ref}' doesn't exist.").format(
                                     secret_name=name, secret_ref=secret_ref))

        found_container_secrets = list(
            filter(lambda cs: cs.secret_id == secret_id and cs.name == name,
                   self.container.container_secrets)
        )

        if not found_container_secrets:
            pecan.abort(404, u._('Secret provided is not in the container'))

        for container_secret in found_container_secrets:
            self.container_secret_repo.delete_entity_by_id(
                container_secret.id, external_project_id)

        pecan.response.status = 204
        LOG.info('Deleted container secret for project: %s',
                 external_project_id)



# Copyright (c) 2013-2014 Rackspace, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
SSL Certificate resources for Barbican.

The resources here should be generic across all certificate-related
implementations. Hence do not place vendor-specific content in this module.
"""

import abc
import datetime

from oslo_config import cfg
from oslo_utils import encodeutils
import six
from stevedore import named

from barbican.common import config
from barbican.common import exception
import barbican.common.utils as utils
from barbican import i18n as u
from barbican.model import models
from barbican.model import repositories as repos
from barbican.plugin.util import utils as plugin_utils

LOG = utils.getLogger(__name__)
CONF = config.new_config()

# Configuration for certificate processing plugins:
DEFAULT_PLUGIN_NAMESPACE = 'barbican.certificate.plugin'
DEFAULT_PLUGINS = ['simple_certificate']

cert_opt_group = cfg.OptGroup(name='certificate',
                              title='Certificate Plugin Options')
cert_opts = [
    cfg.StrOpt('namespace',
               default=DEFAULT_PLUGIN_NAMESPACE,
               help=u._('Extension namespace to search for plugins.')
               ),
    cfg.MultiStrOpt('enabled_certificate_plugins',
                    default=DEFAULT_PLUGINS,
                    help=u._('List of certificate plugins to load.')
                    )
]
CONF.register_group(cert_opt_group)
CONF.register_opts(cert_opts, group=cert_opt_group)
config.parse_args(CONF)


def list_opts():
    yield cert_opt_group, cert_opts
    yield cert_event_opt_group, cert_event_opts


# Configuration for certificate eventing plugins:
DEFAULT_EVENT_PLUGIN_NAMESPACE = 'barbican.certificate.event.plugin'
DEFAULT_EVENT_PLUGINS = ['simple_certificate_event']

cert_event_opt_group = cfg.OptGroup(name='certificate_event',
                                    title='Certificate Event Plugin Options')
cert_event_opts = [
    cfg.StrOpt('namespace',
               default=DEFAULT_EVENT_PLUGIN_NAMESPACE,
               help=u._('Extension namespace to search for eventing plugins.')
               ),
    cfg.MultiStrOpt('enabled_certificate_event_plugins',
                    default=DEFAULT_EVENT_PLUGINS,
                    help=u._('List of certificate plugins to load.')
                    )
]
CONF.register_group(cert_event_opt_group)
CONF.register_opts(cert_event_opts, group=cert_event_opt_group)


ERROR_RETRY_MSEC = 300000
RETRY_MSEC = 3600000
CA_INFO_DEFAULT_EXPIRATION_DAYS = 1

CA_PLUGIN_TYPE_DOGTAG = "dogtag"
CA_PLUGIN_TYPE_SYMANTEC = "symantec"

# fields to distinguish CA types and subject key identifiers
CA_TYPE = "ca_type"
CA_SUBJECT_KEY_IDENTIFIER = "ca_subject_key_identifier"

# field to get the certificate request type
REQUEST_TYPE = "request_type"

# fields for the ca_id, plugin_ca_id
CA_ID = "ca_id"
PLUGIN_CA_ID = "plugin_ca_id"

# fields for ca_info dict keys
INFO_NAME = "name"
INFO_DESCRIPTION = "description"
INFO_CA_SIGNING_CERT = "ca_signing_certificate"
INFO_INTERMEDIATES = "intermediates"
INFO_EXPIRATION = "expiration"


# Singleton to avoid loading the CertificateEventManager plugins more than once
_EVENT_PLUGIN_MANAGER = None


class CertificateRequestType(object):
    """Constants to define the certificate request type."""
    CUSTOM_REQUEST = "custom"
    FULL_CMC_REQUEST = "full-cmc"
    SIMPLE_CMC_REQUEST = "simple-cmc"
    STORED_KEY_REQUEST = "stored-key"


class CertificatePluginNotFound(exception.BarbicanException):
    """Raised when no certificate plugin supporting a request is available."""
    def __init__(self, plugin_name=None):
        if plugin_name:
            message = u._(
                'Certificate plugin "{name}"'
                ' not found.').format(name=plugin_name)
        else:
            message = u._("Certificate plugin not found or configured.")
        super(CertificatePluginNotFound, self).__init__(message)


class CertificatePluginNotFoundForCAID(exception.BarbicanException):
    """Raised when no certificate plugin is available for a CA_ID."""
    def __init__(self, ca_id):
        message = u._(
            'Certificate plugin not found for "{ca_id}".').format(ca_id=ca_id)
        super(CertificatePluginNotFoundForCAID, self).__init__(message)


class CertificateEventPluginNotFound(exception.BarbicanException):
    """Raised with no certificate event plugin supporting request."""
    def __init__(self, plugin_name=None):
        if plugin_name:
            message = u._(
                'Certificate event plugin "{name}" '
                'not found.').format(name=plugin_name)
        else:
            message = u._("Certificate event plugin not found.")
        super(CertificateEventPluginNotFound, self).__init__(message)


class CertificateStatusNotSupported(exception.BarbicanException):
    """Raised when cert status returned is unknown."""
    def __init__(self, status):
        super(CertificateStatusNotSupported, self).__init__(
            u._("Certificate status of {status} not "
                "supported").format(status=status)
        )
        self.status = status


class CertificateGeneralException(exception.BarbicanException):
    """Raised when a system fault has occurred."""
    def __init__(self, reason=u._('Unknown')):
        super(CertificateGeneralException, self).__init__(
            u._('Problem seen during certificate processing - '
                'Reason: {reason}').format(reason=reason)
        )
        self.reason = reason


class CertificateStatusClientDataIssue(exception.BarbicanHTTPException):
    """Raised when the CA has encountered an issue with request data."""

    client_message = ""
    status_code = 400

    def __init__(self, reason=u._('Unknown')):
        super(CertificateStatusClientDataIssue, self).__init__(
            u._('Problem with data in certificate request - '
                'Reason: {reason}').format(reason=reason)
        )
        self.client_message = self.message


class CertificateStatusInvalidOperation(exception.BarbicanHTTPException):
    """Raised when the CA has encountered an issue with request data."""

    client_message = ""
    status_code = 400

    def __init__(self, reason=u._('Unknown')):
        super(CertificateStatusInvalidOperation, self).__init__(
            u._('Invalid operation requested - '
                'Reason: {reason}').format(reason=reason)
        )
        self.client_message = self.message


@six.add_metaclass(abc.ABCMeta)
class CertificateEventPluginBase(object):
    """Base class for certificate eventing plugins.

    This class is the base plugin contract for issuing certificate related
    events from Barbican.
    """

    @abc.abstractmethod
    def notify_certificate_is_ready(
            self, project_id, order_ref, container_ref):
        """Notify that a certificate has been generated and is ready to use.

        :param project_id: Project ID associated with this certificate
        :param order_ref: HATEOAS reference URI to the submitted Barbican Order
        :param container_ref: HATEOAS reference URI to the Container storing
               the certificate
        :returns: None
        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def notify_ca_is_unavailable(
            self, project_id, order_ref, error_msg, retry_in_msec):
        """Notify that the certificate authority (CA) isn't available.

        :param project_id: Project ID associated with this order
        :param order_ref: HATEOAS reference URI to the submitted Barbican Order
        :param error_msg: Error message if it is available
        :param retry_in_msec: Delay before attempting to talk to the CA again.
               If this is 0, then no attempt will be made.
        :returns: None
        """
        raise NotImplementedError  # pragma: no cover


@six.add_metaclass(abc.ABCMeta)
class CertificatePluginBase(object):
    """Base class for certificate plugins.

    This class is the base plugin contract for certificates.
    """

    @abc.abstractmethod
    def get_default_ca_name(self):
        """Get the default CA name

        Provides a default CA name to be returned in the default
        get_ca_info() method.  If get_ca_info() is overridden (to
        support multiple CAs for instance), then this method may not
        be called.  In that case, just implement this method to return
        a dummy variable.

        If this value is used, it should be unique amongst all the CA
        plugins.

        :return: The default CA name
        :rtype: str
        """
        raise NotImplementedError   # pragma: no cover

    @abc.abstractmethod
    def get_default_signing_cert(self):
        """Get the default CA signing cert

        Provides a default CA signing cert to be returned in the default
        get_ca_info() method.  If get_ca_info() is overridden (to
        support multiple CAs for instance), then this method may not
        be called.  In that case, just implement this method to return
        a dummy variable.
        :return: The default CA signing cert
        :rtype: str
        """
        raise NotImplementedError   # pragma: no cover

    @abc.abstractmethod
    def get_default_intermediates(self):
        """Get the default CA certificate chain

        Provides a default CA certificate to be returned in the default
        get_ca_info() method.  If get_ca_info() is overridden (to
        support multiple CAs for instance), then this method may not
        be called.  In that case, just implement this method to return
        a dummy variable.
        :return: The default CA certificate chain
        :rtype: str
        """
        raise NotImplementedError   # pragma: no cover

    @abc.abstractmethod
    def issue_certificate_request(self, order_id, order_meta, plugin_meta,
                                  barbican_meta_dto):
        """Create the initial order

        :param order_id: ID associated with the order
        :param order_meta: Dict of meta-data associated with the order
        :param plugin_meta: Plugin meta-data previously set by calls to
                            this plugin. Plugins may also update/add
                            information here which Barbican will persist
                            on their behalf
        :param barbican_meta_dto:
            Data transfer object :class:`BarbicanMetaDTO` containing data
            added to the request by the Barbican server to provide additional
            context for processing, but which are not in
            the original request.  For example, the plugin_ca_id
        :returns: A :class:`ResultDTO` instance containing the result
                  populated by the plugin implementation
        :rtype: :class:`ResultDTO`
        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def modify_certificate_request(self, order_id, order_meta, plugin_meta,
                                   barbican_meta_dto):
        """Update the order meta-data

        :param order_id: ID associated with the order
        :param order_meta: Dict of meta-data associated with the order
        :param plugin_meta: Plugin meta-data previously set by calls to
                            this plugin. Plugins may also update/add
                            information here which Barbican will persist
                            on their behalf
        :param barbican_meta_dto:
            Data transfer object :class:`BarbicanMetaDTO` containing data
            added to the request by the Barbican server to provide additional
            context for processing, but which are not in
            the original request.  For example, the plugin_ca_id
        :returns: A :class:`ResultDTO` instance containing the result
                  populated by the plugin implementation
        :rtype: :class:`ResultDTO`
        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def cancel_certificate_request(self, order_id, order_meta, plugin_meta,
                                   barbican_meta_dto):
        """Cancel the order

        :param order_id: ID associated with the order
        :param order_meta: Dict of meta-data associated with the order.
        :param plugin_meta: Plugin meta-data previously set by calls to
                            this plugin. Plugins may also update/add
                            information here which Barbican will persist
                            on their behalf
        :param barbican_meta_dto:
            Data transfer object :class:`BarbicanMetaDTO` containing data
            added to the request by the Barbican server to provide additional
            context for processing, but which are not in
            the original request.  For example, the plugin_ca_id
        :returns: A :class:`ResultDTO` instance containing the result
                  populated by the plugin implementation
        :rtype: :class:`ResultDTO`
        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def check_certificate_status(self, order_id, order_meta, plugin_meta,
                                 barbican_meta_dto):
        """Check status of the order

        :param order_id: ID associated with the order
        :param order_meta: Dict of meta-data associated with the order
        :param plugin_meta: Plugin meta-data previously set by calls to
                            this plugin. Plugins may also update/add
                            information here which Barbican will persist
                            on their behalf
        :param barbican_meta_dto:
            Data transfer object :class:`BarbicanMetaDTO` containing data
            added to the request by the Barbican server to provide additional
            context for processing, but which are not in
            the original request.  For example, the plugin_ca_id
        :returns: A :class:`ResultDTO` instance containing the result
                  populated by the plugin implementation
        :rtype: :class:`ResultDTO`
        """
        raise NotImplementedError  # pragma: no cover

    @abc.abstractmethod
    def supports(self, certificate_spec):
        """Returns if the plugin supports the certificate type.

        :param certificate_spec: Contains details on the certificate to
                                 generate the certificate order
        :returns: boolean indicating if the plugin supports the certificate
                  type
        """
        raise NotImplementedError  # pragma: no cover

    def supported_request_types(self):
        """Returns the request_types supported by this plugin.

        :returns: a list of the Barbican-core defined request_types
                  supported by this plugin.
        """
        return [CertificateRequestType.CUSTOM_REQUEST]  # pragma: no cover

    def get_ca_info(self):
        """Returns information about the CA(s) supported by this plugin.

        :returns: dictionary indexed by plugin_ca_id.  Each entry consists
                  of a dictionary of key-value pairs.

        An example dictionary containing the current supported attributes
        is shown below::

            { "plugin_ca_id1": {
                INFO_NAME : "CA name",
                INFO_DESCRIPTION : "CA user friendly description",
                INFO_CA_SIGNING_CERT : "base 64 encoded signing cert",
                INFO_INTERMEDIATES = "base 64 encoded certificate chain"
                INFO_EXPIRATION = "ISO formatted UTC datetime for when this"
                                  "data will become stale"
                }
            }

        """
        name = self.get_default_ca_name()
        expiration = (datetime.datetime.utcnow() +
                      datetime.timedelta(days=CA_INFO_DEFAULT_EXPIRATION_DAYS))

        default_info = {
            INFO_NAME: name,
            INFO_DESCRIPTION: "Certificate Authority - {0}".format(name),
            INFO_EXPIRATION: expiration.isoformat()
        }

        signing_cert = self.get_default_signing_cert()
        if signing_cert is not None:
            default_info[INFO_CA_SIGNING_CERT] = signing_cert

        intermediates = self.get_default_intermediates()
        if intermediates is not None:
            default_info[INFO_INTERMEDIATES] = intermediates

        return {name: default_info}

    def supports_create_ca(self):
        """Returns whether the plugin supports on-the-fly generation of subCAs

        :return: boolean, True if supported, defaults to False
        """
        return False    # pragma: no cover

    def create_ca(self, ca_create_dto):
        """Creates a subordinate CA upon request

        This call should only be made if a plugin returns True for
        supports_create_ca().

        :param ca_create_dto:
            Data transfer object :class:`CACreateDTO` containing data
            required to generate a subordinate CA.  This data includes
            the subject DN of the new CA signing certificate, a name for
            the new CA and a reference to the CA that will issue the new
            subordinate CA's signing certificate,

        :return: ca_info:
            Dictionary containing the data needed to create a
            models.CertificateAuthority object
        """
        raise NotImplementedError    # pragma: no cover

    def delete_ca(self, ca_id):
        """Deletes a subordinate CA

        Like the create_ca call, this should only be made if the plugin
        returns True for supports_create_ca()

        :param ca_id: id for the CA as specified by the plugin
        :return: None
        """
        raise NotImplementedError   # pragma: no cover


class CACreateDTO(object):
    """Class that includes data needed to create a subordinate CA """

    def __init__(self, name=None, description=None, subject_dn=None,
                 parent_ca_id=None):
        """Creates a new CACreateDTO object.

        :param name: Name for the  subordinate CA
        :param description: Description for the subordinate CA
        :param subject_dn:
            Subject DN for the new subordinate CA's signing certificate
        :param parent_ca_id:
            ID of the CA which is supposed to sign the subordinate CA's
            signing certificate.  This is ID as known to the plugin
            (not the Barbican UUID)
        """
        self.name = name
        self.description = description
        self.subject_dn = subject_dn
        self.parent_ca_id = parent_ca_id


class CertificateStatus(object):
    """Defines statuses for certificate request process.

    In particular:

    CERTIFICATE_GENERATED - Indicates a certificate was created

    WAITING_FOR_CA - Waiting for Certificate authority (CA) to complete order

    CLIENT_DATA_ISSUE_SEEN - Problem was seen with client-provided data

    CA_UNAVAILABLE_FOR_REQUEST - CA was not available, will try again later

    REQUEST_CANCELED - The client or CA cancelled this order

    INVALID_OPERATION - Unexpected error seen processing order
    """

    CERTIFICATE_GENERATED = "certificate generated"
    WAITING_FOR_CA = "waiting for CA"
    CLIENT_DATA_ISSUE_SEEN = "client data issue seen"
    CA_UNAVAILABLE_FOR_REQUEST = "CA unavailable for request"
    REQUEST_CANCELED = "request canceled"
    INVALID_OPERATION = "invalid operation"


class ResultDTO(object):
    """Result data transfer object (DTO).

    An object of this type is returned by most certificate plugin methods, and
    is used to guide follow on processing and to provide status feedback to
    clients.
    """
    def __init__(self, status, status_message=None, certificate=None,
                 intermediates=None, retry_msec=RETRY_MSEC, retry_method=None):
        """Creates a new ResultDTO.

        :param status: Status for cert order
        :param status_message: Message to explain status type.
        :param certificate: Certificate returned from CA to be stored in
                            container
        :param intermediates: Intermediates to be stored in container
        :param retry_msec: Number of milliseconds to wait for retry
        :param retry_method: Method to be called for retry, if None then retry
                             the current method
        """
        self.status = status
        self.status_message = status_message
        self.certificate = certificate
        self.intermediates = intermediates
        self.retry_msec = int(retry_msec)
        self.retry_method = retry_method


class BarbicanMetaDTO(object):
    """Barbican meta data transfer object

    Information needed to process a certificate request that is not specified
    in the original request, and written by Barbican core, that is needed
    by the plugin to process requests.
    """

    def __init__(self, plugin_ca_id=None, generated_csr=None):
        """Creates a new BarbicanMetaDTO.

        :param plugin_ca_id: ca_id as known to the plugin
        :param generated_csr: csr generated in the stored-key case
        :return: BarbicanMetaDTO
        """
        self.plugin_ca_id = plugin_ca_id
        self.generated_csr = generated_csr


class CertificatePluginManager(named.NamedExtensionManager):
    def __init__(self, conf=CONF, invoke_args=(), invoke_kwargs={}):
        self.ca_repo = repos.get_ca_repository()
        super(CertificatePluginManager, self).__init__(
            conf.certificate.namespace,
            conf.certificate.enabled_certificate_plugins,
            invoke_on_load=False,  # Defer creating plugins to utility below.
            invoke_args=invoke_args,
            invoke_kwds=invoke_kwargs
        )

        plugin_utils.instantiate_plugins(
            self, invoke_args, invoke_kwargs)

    def get_plugin(self, certificate_spec):
        """Gets a supporting certificate plugin.

        :param certificate_spec: Contains details on the certificate to
                                 generate the certificate order
        :returns: CertificatePluginBase plugin implementation
        """
        request_type = certificate_spec.get(
            REQUEST_TYPE,
            CertificateRequestType.CUSTOM_REQUEST)

        for plugin in plugin_utils.get_active_plugins(self):
            supported_request_types = plugin.supported_request_types()
            if request_type not in supported_request_types:
                continue

            if plugin.supports(certificate_spec):
                return plugin

        raise CertificatePluginNotFound()

    def get_plugin_by_name(self, plugin_name):
        """Gets a supporting certificate plugin.

        :param plugin_name: Name of the plugin to invoke
        :returns: CertificatePluginBase plugin implementation
        """
        for plugin in plugin_utils.get_active_plugins(self):
            if utils.generate_fullname_for(plugin) == plugin_name:
                return plugin
        raise CertificatePluginNotFound(plugin_name)

    def get_plugin_by_ca_id(self, ca_id):
        """Gets a plugin based on the ca_id.

        :param ca_id: id for CA in the CertificateAuthorities table
        :returns: CertificatePluginBase plugin implementation
        """
        ca = self.ca_repo.get(ca_id, suppress_exception=True)
        if not ca:
            raise CertificatePluginNotFoundForCAID(ca_id)

        return self.get_plugin_by_name(ca.plugin_name)

    def refresh_ca_table(self):
        """Refreshes the CertificateAuthority table."""
        updates_made = False
        for plugin in plugin_utils.get_active_plugins(self):
            plugin_name = utils.generate_fullname_for(plugin)
            cas, offset, limit, total = self.ca_repo.get_by_create_date(
                plugin_name=plugin_name,
                suppress_exception=True)
            if total < 1:
                # if no entries are found, then the plugin has not yet been
                # queried or that plugin's entries have expired.
                # Most of the time, this will be a no-op for plugins.
                self.update_ca_info(plugin)
                updates_made = True
        if updates_made:
            # commit to DB to avoid async issues with different threads
            repos.commit()

    def update_ca_info(self, cert_plugin):
        """Update the CA info for a particular plugin."""

        plugin_name = utils.generate_fullname_for(cert_plugin)
        try:
            new_ca_infos = cert_plugin.get_ca_info()
        except Exception as e:
            # The plugin gave an invalid CA, log and return
            LOG.error("ERROR getting CA from plugin: %s",
                      encodeutils.exception_to_unicode(e))
            return

        old_cas, offset, limit, total = self.ca_repo.get_by_create_date(
            plugin_name=plugin_name,
            suppress_exception=True,
            show_expired=True)

        if old_cas:
            for old_ca in old_cas:
                plugin_ca_id = old_ca.plugin_ca_id
                if plugin_ca_id not in new_ca_infos.keys():
                    # remove CAs that no longer exist
                    self._delete_ca(old_ca)
                else:
                    # update those that still exist
                    self.ca_repo.update_entity(
                        old_ca,
                        new_ca_infos[plugin_ca_id])
            old_ids = set([ca.plugin_ca_id for ca in old_cas])
        else:
            old_ids = set()

        new_ids = set(new_ca_infos.keys())

        # add new CAs
        add_ids = new_ids - old_ids
        for add_id in add_ids:
            try:
                self._add_ca(plugin_name, add_id, new_ca_infos[add_id])
            except Exception as e:
                # The plugin gave an invalid CA, log and continue
                LOG.error("ERROR adding CA from plugin: %s",
                          encodeutils.exception_to_unicode(e))

    def _add_ca(self, plugin_name, plugin_ca_id, ca_info):
        parsed_ca = dict(ca_info)
        parsed_ca['plugin_name'] = plugin_name
        parsed_ca['plugin_ca_id'] = plugin_ca_id
        new_ca = models.CertificateAuthority(parsed_ca)
        self.ca_repo.create_from(new_ca)

    def _delete_ca(self, ca):
        self.ca_repo.delete_entity_by_id(ca.id, None)


class _CertificateEventPluginManager(named.NamedExtensionManager,
                                     CertificateEventPluginBase):
    """Provides services for certificate event plugins.

    This plugin manager differs from others in that it implements the same
    contract as the plugins that it manages. This allows eventing operations
    to occur on all installed plugins (with this class acting as a composite
    plugin), rather than just eventing via an individual plugin.

    Each time this class is initialized it will load a new instance
    of each enabled plugin. This is undesirable, so rather than initializing a
    new instance of this class use the get_event_plugin_manager function
    at the module level.
    """
    def __init__(self, conf=CONF, invoke_args=(), invoke_kwargs={}):
        super(_CertificateEventPluginManager, self).__init__(
            conf.certificate_event.namespace,
            conf.certificate_event.enabled_certificate_event_plugins,
            invoke_on_load=False,  # Defer creating plugins to utility below.
            invoke_args=invoke_args,
            invoke_kwds=invoke_kwargs
        )

        plugin_utils.instantiate_plugins(
            self, invoke_args, invoke_kwargs)

    def get_plugin_by_name(self, plugin_name):
        """Gets a supporting certificate event plugin.

        :returns: CertificateEventPluginBase plugin implementation
        """
        for plugin in plugin_utils.get_active_plugins(self):
            if utils.generate_fullname_for(plugin) == plugin_name:
                return plugin
        raise CertificateEventPluginNotFound(plugin_name)

    def notify_certificate_is_ready(
            self, project_id, order_ref, container_ref):
        self._invoke_certificate_plugins(
            'notify_certificate_is_ready',
            project_id, order_ref, container_ref)

    def notify_ca_is_unavailable(
            self, project_id, order_ref, error_msg, retry_in_msec):
        self._invoke_certificate_plugins(
            'notify_ca_is_unavailable',
            project_id, order_ref, error_msg, retry_in_msec)

    def _invoke_certificate_plugins(self, method, *args, **kwargs):
        """Invoke same function on plugins as calling function."""
        active_plugins = plugin_utils.get_active_plugins(self)

        if not active_plugins:
            raise CertificateEventPluginNotFound()

        for plugin in active_plugins:
            getattr(plugin, method)(*args, **kwargs)


def get_event_plugin_manager():
    global _EVENT_PLUGIN_MANAGER
    if _EVENT_PLUGIN_MANAGER:
        return _EVENT_PLUGIN_MANAGER
    _EVENT_PLUGIN_MANAGER = _CertificateEventPluginManager()
    return _EVENT_PLUGIN_MANAGER
